import numpy as np

from . import BaseAnnotationDefinition


class Polyline(BaseAnnotationDefinition):
    """
    Polyline annotation object
    """

    def __init__(self, geo, label, attributes=None):

        self.type = "polyline"
        self.is_open = True
        self.geo = geo
        self.label = label
        if attributes is None:
            attributes = list()
        self.attributes = attributes

    @property
    def x(self):
        return self.geo[:, 0]

    @property
    def y(self):
        return self.geo[:, 1]

    @property
    def left(self):
        return np.min(self.x)

    @property
    def top(self):
        return np.min(self.y)

    @property
    def right(self):
        return np.max(self.x)

    @property
    def bottom(self):
        return np.max(self.y)

    def to_coordinates(self, color):
        return [{"x": float(x), "y": float(y)} for x, y in self.geo]

    @staticmethod
    def from_coordinates(coordinates):
        return np.asarray([[pt["x"], pt["y"]] for pt in coordinates])

    def show(self, image, thickness, with_text, height, width, annotation_format, color):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: dl.ViewAnnotationOptions.list()
        :param color: color
        :return: ndarray
        """
        try:
            import cv2
        except ImportError:
            self.logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        # polyline cant have thickness -1
        if thickness is None or thickness == -1:
            thickness = 1

        # draw annotation
        image = cv2.polylines(
            img=image,
            pts=[np.round(self.geo).astype(int)],
            color=color,
            isClosed=False,
            thickness=thickness,
        )
        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            geo = cls.from_coordinates(coordinates=_json["coordinates"])
        elif "data" in _json:
            geo = cls.from_coordinates(coordinates=_json["data"])
        else:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))
        attributes = _json.get("attributes", list())
        return cls(
            geo=geo,
            label=_json["label"],
            attributes=attributes,
        )



from collections import namedtuple
import logging
import random
import attr

from .. import entities, PlatformException, repositories, services

logger = logging.getLogger(name=__name__)


@attr.s
class Ontology(entities.BaseEntity):
    """
    Ontology object
    """
    # api
    _client_api = attr.ib(type=services.ApiClient, repr=False)

    # params
    id = attr.ib()
    creator = attr.ib()
    url = attr.ib(repr=False)
    labels = attr.ib()
    metadata = attr.ib(repr=False)
    attributes = attr.ib()

    # entities
    _recipe = attr.ib(repr=False)

    # repositories
    _repositories = attr.ib(repr=False)

    # defaults
    _instance_map = attr.ib(default=None, repr=False)

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['ontologies'])

        if self._recipe is None:
            ontologies = repositories.Ontologies(client_api=self._client_api, recipe=self._recipe)
        else:
            ontologies = self.recipe.ontologies

        r = reps(ontologies=ontologies)
        return r

    @property
    def recipe(self):
        assert isinstance(self._recipe, entities.Recipe)
        return self._recipe

    @property
    def ontologies(self):
        assert isinstance(self._repositories.ontologies, repositories.Ontologies)
        return self._repositories.ontologies

    @property
    def labels_flat_dict(self):
        flatten_dict = dict()

        def add_to_dict(tag, father):
            flatten_dict[tag] = father
            for child in father.children:
                add_to_dict('{}.{}'.format(tag, child.tag), child)

        for label in self.labels:
            add_to_dict(label.tag, label)
        return flatten_dict

    @property
    def instance_map(self):
        if self._instance_map is None:
            labels = [label for label in self.labels_flat_dict]
            labels.sort()
            # each label gets index as instance id
            self._instance_map = {label: (i_label + 1) for i_label, label in enumerate(labels)}
        return self._instance_map

    @classmethod
    def from_json(cls, _json, client_api, recipe):
        """
        Build an Ontology entity object from a json

        :param _json: _json response from host
        :param recipe: ontology's recipe
        :param client_api: client_api
        :return: Ontology object
        """
        if "attributes" in _json:
            attributes = _json["attributes"]
        else:
            attributes = list()

        labels = list()
        for root in _json["roots"]:
            labels.append(entities.Label.from_root(root=root))

        return cls(
            metadata=_json.get("metadata", None),
            creator=_json.get("creator", None),
            url=_json.get("url", None),
            id=_json["id"],
            attributes=attributes,
            client_api=client_api,
            recipe=recipe,
            labels=labels,
        )

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        roots = [label.to_root() for label in self.labels]
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Ontology)._client_api,
                                                              attr.fields(Ontology)._recipe,
                                                              attr.fields(Ontology)._instance_map,
                                                              attr.fields(Ontology)._repositories))
        _json["roots"] = roots
        return _json

    def delete(self):
        """
        Delete recipe from platform

        :return: True
        """
        return self.ontologies.delete(self.id)

    def update(self, system_metadata=False):
        """
        Update items metadata

        :param system_metadata: bool
        :return: Ontology object
        """
        return self.ontologies.update(self, system_metadata=system_metadata)

    def add_label(self, label_name, color=None, children=None, attributes=None, display_label=None, label=None,
                  add=True):
        """
        Add a single label to ontology

        :param add:
        :param label:
        :param label_name: label name
        :param color: optional - if not given a random color will be selected
        :param children: optional - children
        :param attributes: optional - attributes
        :param display_label: optional - display_label
        :return: Label entity
        """
        if not isinstance(label, entities.Label):
            if attributes is None:
                attributes = list()
            if not isinstance(attributes, list):
                attributes = [attributes]

            # get random color if none given
            if color is None:
                color = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

            if children is None:
                children = list()
            if not isinstance(children, list):
                children = [children]

            # add children    
            added_children = list()
            for child in children:
                if not isinstance(child, entities.Label):
                    added_children.append(self.add_label(**child, add=False))
                else:
                    added_children.append(child)

            if display_label is None:
                display_label = ""
                if len(label_name.split("_")) == 1:
                    display_label = label_name[0].upper() + label_name[1:]
                else:
                    for word in label_name.split("_"):
                        display_label += word[0].upper() + word[1:] + " "
                    display_label = display_label[0:-1]
            root = {
                "value": {
                    "tag": label_name,
                    "displayLabel": display_label,
                    "color": color,
                    "attributes": attributes,
                },
                "children": list(),
            }
            added_label = entities.Label.from_root(root)
            added_label.children = added_children
        else:
            added_label = label

        if add:
            self.labels.append(added_label)
        return added_label

    def add_labels(self, label_list):
        """
        Adds a list of labels to ontology

        :param label_list: list of labels [{"value": {"tag": "tag", "displayLabel": "displayLabel",
                                            "color": "#color", "attributes": [attributes]}, "children": [children]}]
        :return: List of label entities added
        """
        labels = list()
        for label in label_list:

            if isinstance(label, str):
                label = entities.Label(tag=label)

            if isinstance(label, entities.Label):
                # label entity
                labels.append(
                    {
                        "label_name": label.tag,
                        "color": label.color,
                        "children": label.children,
                        "attributes": label.attributes,
                        "display_label": label.display_label,
                    }
                )
            else:
                # dictionary
                children = label.get("children", None)
                label = label.get("value", label)
                if "tag" in label:
                    label_name = label["tag"]
                elif "label_name" in label:
                    label_name = label["label_name"]
                else:
                    raise PlatformException("400", "Invalid input - each label must have a tag")
                if "color" in label:
                    color = label["color"]
                else:
                    logger.warning('No color given for label: {}, random color will be selected'.format(label_name))
                    color = None
                attributes = label.get("attributes", None)
                display_label = label.get("displayLabel", None)
                labels.append(
                    {
                        "label_name": label_name,
                        "color": color,
                        "children": children,
                        "attributes": attributes,
                        "display_label": display_label,
                    }
                )

        added_labels = list()
        for label in labels:
            added_labels.append(self.add_label(label_name=label["label_name"],
                                               color=label["color"],
                                               children=label["children"],
                                               attributes=label["attributes"],
                                               display_label=label["display_label"]))

        return added_labels

    def delete_labels(self, label_names):
        """
        Delete labels from ontology

        :param label_names: label object/ label name / list of label objects / list of label names
        :return:
        """
        if not isinstance(label_names, list):
            label_names = [label_names]

        if isinstance(label_names[0], entities.Label):
            label_names = [label.tag for label in label_names]

        for label in label_names:
            self.__delete_label(label)

        self.update()

    def __delete_label(self, label_name):
        if label_name in self.instance_map.keys():
            labels = self.labels
            label_chain = label_name.split('.')
            while len(label_chain) > 1:
                label_name = label_chain.pop(0)
                for i_label, label in enumerate(labels):
                    if label.tag == label_name:
                        labels = labels[i_label].children
                        break
            label_name = label_chain[0]
            for i_label, label in enumerate(labels):
                if label.tag == label_name:
                    labels.pop(i_label)



import traceback
from collections import namedtuple
import logging
import attr

from .. import repositories, entities, services

logger = logging.getLogger(name=__name__)


@attr.s
class Recipe(entities.BaseEntity):
    """
    Recipe object
    """
    id = attr.ib()
    creator = attr.ib()
    url = attr.ib(repr=False)
    title = attr.ib()
    project_ids = attr.ib()
    description = attr.ib()
    ontologyIds = attr.ib(repr=False)
    instructions = attr.ib(repr=False)
    examples = attr.ib(repr=False)
    customActions = attr.ib(repr=False)
    metadata = attr.ib()

    # name change
    ui_settings = attr.ib()

    # platform
    _client_api = attr.ib(type=services.ApiClient, repr=False)
    # entities
    _dataset = attr.ib(repr=False, default=None)
    _project = attr.ib(repr=False, default=None)
    # repositories
    _repositories = attr.ib(repr=False)

    @classmethod
    def from_json(cls, _json, client_api, dataset=None, project=None, is_fetched=True):
        """
        Build a Recipe entity object from a json

        :param _json: _json response from host
        :param dataset: recipe's dataset
        :param project: recipe's project
        :param client_api: client_api
        :param is_fetched: is Entity fetched from Platform
        :return: Recipe object
        """
        inst = cls(
            client_api=client_api,
            dataset=dataset,
            project=project,
            id=_json['id'],
            creator=_json.get('creator', None),
            url=_json.get('url', None),
            title=_json.get('title', None),
            project_ids=_json.get('projectIds', None),
            description=_json.get('description', None),
            ontologyIds=_json.get('ontologyIds', None),
            instructions=_json.get('instructions', None),
            ui_settings=_json.get('uiSettings', None),
            metadata=_json.get('metadata', None),
            examples=_json.get('examples', None),
            customActions=_json.get('customActions', None))
        inst.is_fetched = is_fetched
        return inst

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, dataset=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
        :param _json:
        :param client_api:
        :param dataset:
        :return:
        """
        try:
            recipe = Recipe.from_json(_json=_json,
                                      client_api=client_api,
                                      project=project,
                                      dataset=dataset,
                                      is_fetched=is_fetched)
            status = True
        except Exception:
            recipe = traceback.format_exc()
            status = False
        return status, recipe

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['ontologies', 'recipes'])
        if self._dataset is None and self._project is None:
            recipes = repositories.Recipes(client_api=self._client_api, dataset=self._dataset, project=self._project)
        elif self._dataset is not None:
            recipes = self.dataset.recipes
        else:
            recipes = self.project.recipes
        r = reps(ontologies=repositories.Ontologies(recipe=self, client_api=self._client_api),
                 recipes=recipes)
        return r

    @property
    def dataset(self):
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @property
    def project(self):
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def recipes(self):
        assert isinstance(self._repositories.recipes, repositories.Recipes)
        return self._repositories.recipes

    @property
    def ontologies(self):
        assert isinstance(self._repositories.ontologies, repositories.Ontologies)
        return self._repositories.ontologies

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Recipe)._client_api,
                                                              attr.fields(Recipe)._dataset,
                                                              attr.fields(Recipe)._project,
                                                              attr.fields(Recipe).project_ids,
                                                              attr.fields(Recipe).ui_settings,
                                                              attr.fields(Recipe)._repositories))
        _json['uiSettings'] = self.ui_settings
        _json['projectIds'] = self.project_ids
        return _json

    def delete(self):
        """
        Delete recipe from platform

        :return: True
        """
        return self.recipes.delete(self.id)

    def update(self, system_metadata=False):
        """
        Update Recipe

        :param system_metadata: bool
        :return: Recipe object
        """
        return self.recipes.update(recipe=self, system_metadata=system_metadata)



import hashlib
import logging
import os
import io
import random

from .. import entities, PlatformException, exceptions, repositories, miscellaneous, services

logger = logging.getLogger(name=__name__)


class Codebases:
    """
    Codebases repository
    """

    def __init__(self,
                 client_api: services.ApiClient,
                 project: entities.Project = None,
                 dataset: entities.Dataset = None):
        self._client_api = client_api
        if project is None and dataset is None:
            raise PlatformException('400', 'at least one must be not None: dataset or project')
        self._project = project
        self._dataset = dataset
        self._items_repository = None

    @property
    def items_repository(self) -> repositories.Items:
        if self._items_repository is None:
            self._items_repository = self.dataset.items
            self._items_repository.set_items_entity(entities.Codebase)
        assert isinstance(self._items_repository, repositories.Items)
        return self._items_repository

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            self._project = self.dataset.project
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            # get dataset from project
            try:
                self._dataset = self.project.datasets.get(dataset_name='Binaries')
            except exceptions.NotFound:
                self._dataset = None
            if self._dataset is None:
                logger.debug(
                    'Dataset for codebases was not found. Creating... dataset name: "Binaries". project_id={}'.format(
                        self.project.id))
                self._dataset = self.project.datasets.create(dataset_name='Binaries')
                # add system to metadata
                if 'metadata' not in self._dataset.to_json():
                    self._dataset.metadata = dict()
                if 'system' not in self._dataset.metadata:
                    self._dataset.metadata['system'] = dict()
                self._dataset.metadata['system']['scope'] = 'system'
                self.project.datasets.update(dataset=self._dataset, system_metadata=True)
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: entities.Dataset):
        if not isinstance(dataset, entities.Dataset):
            raise ValueError('Must input a valid Dataset entity')
        self._dataset = dataset

    @staticmethod
    def __file_hash(filepath):
        m = hashlib.md5()
        with open(filepath, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b''):
                m.update(chunk)
        return m.hexdigest()

    def list_versions(self, codebase_name):
        """
        List all codebase versions

        :param codebase_name: code base name
        :return: list of versions
        """
        filters = entities.Filters()
        filters.add(field='filename', values='/codebases/{}/*'.format(codebase_name))
        versions = self.items_repository.list(filters=filters)
        return versions

    def list(self) -> entities.PagedEntities:
        """
        List all code bases
        :return: Paged entity
        """
        filters = entities.Filters()
        filters.add(field='filename', values='/codebases/*')
        filters.add(field='type', values='dir')
        codebases = self.items_repository.list(filters=filters)
        return codebases

    def get(self, codebase_name=None, codebase_id=None, version=None):
        """
        Get a Codebase object
        :param version: codebase version. default is latest. options: "all", "latests" or ver number - "10"
        :param codebase_id: optional - search by id
        :param codebase_name: optional - search by name
        :return: Codebase object
        """
        if codebase_id is not None:
            matched_version = self.items_repository.get(item_id=codebase_id)
            return matched_version

        if codebase_name is None:
            raise PlatformException(error='400', message='Either "codebase_name" or "codebase_id" is needed')
        if version is None:
            version = 'latest'

        if version not in ['all', 'latest']:
            try:
                matched_version = self.items_repository.get(
                    filepath='/codebases/{}/{}.zip'.format(codebase_name, version))
            except Exception:
                raise PlatformException(error='404',
                                        message='No matching version was found. version: {}'.format(version))
            return matched_version

        # get all or latest
        versions_pages = self.list_versions(codebase_name=codebase_name)
        if versions_pages.items_count == 0:
            raise PlatformException(error='404', message='No codebase was found. name: {}'.format(codebase_name))
        else:
            if version == 'all':
                matched_version = versions_pages
            elif version == 'latest':
                max_ver = -1
                matched_version = None
                for page in versions_pages:
                    for ver in page:
                        if ver.type == 'dir':
                            continue
                        # extract version from filepath
                        ver_int = int(os.path.splitext(ver.name)[0])
                        if ver_int > max_ver:
                            max_ver = ver_int
                            matched_version = ver
                if matched_version is None:
                    raise PlatformException(error='404',
                                            message='No codebase was found. name: {}'.format(codebase_name))
            else:
                raise PlatformException(error='404', message='Unknown version string: {}'.format(version))

        return matched_version

    @staticmethod
    def get_current_version(all_versions_pages, zip_md):
        latest_version = 0
        same_version_found = None
        # go over all existing versions
        for page in all_versions_pages:
            for v_item in page:
                # get latest version
                if int(os.path.splitext(v_item.name)[0]) > latest_version:
                    latest_version = int(os.path.splitext(v_item.name)[0])
                # check md5 to find same codebase
                if v_item.md5 == zip_md:
                    same_version_found = v_item
                    break
        return latest_version + 1, same_version_found

    def pack(self, directory, name=None, description=''):
        """
        Zip a local code directory and post to codebases
        :param directory: local directory to pack
        :param name: codebase name
        :param description: codebase description
        :return: Codebase object
        """
        # create/get .dataloop dir
        cwd = os.getcwd()
        dl_dir = os.path.join(cwd, '.dataloop')
        if not os.path.isdir(dl_dir):
            os.mkdir(dl_dir)

        # get codebase name
        if name is None:
            name = os.path.basename(directory)

        # create/get dist folder
        zip_filename = os.path.join(dl_dir, '{}_{}.zip'.format(name, str(random.randrange(0, 1000))))

        try:
            if not os.path.isdir(directory):
                raise PlatformException(error='400', message='Not a directory: {}'.format(directory))
            directory = os.path.abspath(directory)

            # create zipfile
            miscellaneous.Zipping.zip_directory(zip_filename=zip_filename, directory=directory)
            zip_md = self.__file_hash(zip_filename)

            # get latest version
            same_version_found = None
            try:
                all_versions_pages = self.get(codebase_name=name, version='all')
            except exceptions.NotFound:
                all_versions_pages = None
            if all_versions_pages is None:
                # no codebase with that name - create new version
                current_version = 0
            else:
                current_version, same_version_found = self.get_current_version(all_versions_pages=all_versions_pages,
                                                                               zip_md=zip_md)

            if same_version_found is not None:
                # same md5 hash file found in version - return the matched version
                item = same_version_found
            else:
                # no matched version was found - create a new version
                # read from zipped file
                with open(zip_filename, 'rb') as f:
                    buffer = io.BytesIO(f.read())
                    buffer.name = str(current_version) + '.zip'

                # upload item
                item = self.items_repository.upload(local_path=buffer,
                                                    remote_path='/codebases/{}'.format(name))
                if isinstance(item, list) and len(item) == 0:
                    raise PlatformException(error='400', message='Failed upload codebase, check log file for details')

                # add source code to metadata
                if 'system' not in item.metadata:
                    item.metadata['system'] = dict()
                item.metadata['system']['description'] = description
                item.metadata['system']['md5'] = zip_md

                # add git info to metadata
                if miscellaneous.GitUtils.is_git_repo(path=directory):
                    # create 'git' field in metadata
                    if 'git' not in item.metadata:
                        item.metadata['git'] = dict()

                    # add to metadata
                    item.metadata['git']['status'] = miscellaneous.GitUtils.git_status(path=directory)
                    item.metadata['git']['log'] = miscellaneous.GitUtils.git_log(path=directory)
                    item.metadata['git']['url'] = miscellaneous.GitUtils.git_url(path=directory)

                # update item
                item = self.items_repository.update(item=item, system_metadata=True)

        except Exception:
            logger.error('Error when packing:')
            raise
        finally:
            # cleanup
            if zip_filename is not None:
                if os.path.isfile(zip_filename):
                    os.remove(zip_filename)
        return item

    def unpack_single(self, codebase: entities.Codebase, download_path, local_path):
        # downloading with specific filename
        artifact_filepath = self.items_repository.download(items=codebase.id,
                                                           save_locally=True,
                                                           local_path=os.path.join(download_path, codebase.name),
                                                           to_items_folder=False)
        if not os.path.isfile(artifact_filepath):
            raise PlatformException(error='404',
                                    message='error downloading codebase. see above for more information')
        miscellaneous.Zipping.unzip_directory(zip_filename=artifact_filepath,
                                              to_directory=local_path)
        os.remove(artifact_filepath)
        return artifact_filepath

    def unpack(self, codebase_name=None, codebase_id=None, local_path=None, version=None):
        """
        Unpack codebase locally. Download source code and unzip
        :param codebase_name: search by name
        :param codebase_id: search by id
        :param local_path: local path to save codebase
        :param version: codebase version to unpack. default - latest
        :return: String (dirpath)
        """
        codebase = self.get(codebase_name=codebase_name,
                            codebase_id=codebase_id,
                            version=version)
        download_path = local_path
        if isinstance(codebase, entities.PagedEntities):
            for page in codebase:
                for item in page:
                    local_path = os.path.join(download_path, 'v.' + item.name.split('.')[0])
                    self.unpack_single(codebase=item,
                                       download_path=download_path,
                                       local_path=local_path)
            return os.path.dirname(local_path)
        elif isinstance(codebase, entities.Codebase):
            artifact_filepath = self.unpack_single(codebase=codebase,
                                                   download_path=download_path,
                                                   local_path=local_path)
            logger.info('Source code was unpacked to: {}'.format(os.path.dirname(artifact_filepath)))
        else:
            raise PlatformException(
                error='404',
                message='Codebase was not found! name:{name}, id:{id}'.format(name=codebase_name,
                                                                              id=codebase_id))
        return os.path.dirname(artifact_filepath)



#! /usr/bin/env python3
# This file is part of DTLPY.
#
# DTLPY is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DTLPY is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DTLPY.  If not, see <http://www.gnu.org/licenses/>.
from .annotations import Annotations
from .artifacts import Artifacts
from .datasets import Datasets
from .items import Items
from .codebases import Codebases
from .projects import Projects
from .ontologies import Ontologies
from .recipes import Recipes
from .packages import Packages
from .downloader import Downloader
from .uploader import Uploader
from .triggers import Triggers
from .times_series import TimesSeries
from .services import Services
from .executions import Executions
from .assignments import Assignments
from .tasks import Tasks
from .bots import Bots
from .webhooks import Webhooks
from .models import Models
from .snapshots import Snapshots
from .analytics import Analytics



class Dtlpy:
    from .exceptions import PlatformException
    from . import repositories, exceptions, entities, examples
    from .__version__ import version as __version__
    from .entities import Box, Point, Segmentation, Polygon, Ellipse, Classification, Subtitle, Polyline, Filters, \
        Trigger, \
        AnnotationCollection, Annotation, Item, Codebase, Filters, Execution, Recipe, Ontology, Label, Similarity, \
        ItemLink, UrlLink, PackageModule, PackageFunction, FunctionIO, Modality, Workload, WorkloadUnit
    from .utilities import Converter, BaseServiceRunner, Progress
    from .services import ApiClient, check_sdk

    def __init__(self, cookie_filepath=None):
        self.client_api = self.ApiClient(cookie_filepath=cookie_filepath)
        self.projects = self.repositories.Projects(client_api=self.client_api)
        self.datasets = self.repositories.Datasets(client_api=self.client_api,
                                                   project=None)
        self.items = self.repositories.Items(client_api=self.client_api,
                                             datasets=self.datasets)
        self.packages = self.repositories.Packages(client_api=self.client_api)
        self.executions = self.repositories.Executions(client_api=self.client_api)
        self.services = self.repositories.Services(client_api=self.client_api)
        self.webhooks = self.repositories.Webhooks(client_api=self.client_api)
        self.triggers = self.repositories.Triggers(client_api=self.client_api)
        self.assignments = self.repositories.Assignments(client_api=self.client_api)
        self.tasks = self.repositories.Tasks(client_api=self.client_api)
        self.annotations = self.repositories.Annotations(client_api=self.client_api)
        self.verbose = self.client_api.verbose
        self.login = self.client_api.login
        self.login_token = self.client_api.login_token
        self.login_secret = self.client_api.login_secret
        self.add_environment = self.client_api.add_environment
        self.setenv = self.client_api.setenv
        self.token_expired = self.client_api.token_expired
        self.info = self.client_api.info

    def __del__(self):
        for name, pool in self.client_api._thread_pools.items():
            pool.close()
            pool.terminate()

    def token(self):
        """
        token
        :return: token in use
        """
        return self.client_api.token

    def environment(self):
        """
        environment
        :return: current environment
        """
        return self.client_api.environment

    def init(self):
        """
        init current directory as a Dataloop working directory
        :return:
        """
        from .services import CookieIO
        self.client_api.state_io = CookieIO.init_local_cookie(create=True)
        assert isinstance(self.client_api.state_io, CookieIO)

    def checkout_state(self):
        """
        Return the current checked out state
        :return:
        """
        state = self.client_api.state_io.read_json()
        return state

    class ModalityTypeEnum:
        """
        State enum
        """
        OVERLAY = 'overlay'

    class SimilarityTypeEnum:
        """
        State enum
        """
        ID = 'id'
        URL = 'url'

    class LinkTypeEnum:
        """
        State enum
        """
        ID = 'id'
        URL = 'url'

    class TriggerResource:
        ITEM = 'Item'
        DATASET = 'Dataset'
        ANNOTATION = 'Annotation'

    class TriggerAction:
        CREATED = 'Created'
        UPDATED = 'Updated'
        DELETED = 'Deleted'

    class TriggerExecutionMode:
        ONCE = 'Once'
        ALWAYS = 'Always'

    class PackageInputType:
        ITEM = 'Item'
        DATASET = 'Dataset'
        ANNOTATION = 'Annotation'
        JSON = 'Json'

    class FiltersResource:
        ITEM = 'items'
        ANNOTATION = 'annotations'

    class FiltersOperations:
        OR = 'or'
        AND = 'and'
        IN = 'in'
        NOT_EQUAL = 'ne'
        EQUAL = 'eq'
        GREATER_THAN = 'gt'
        GLOB = 'glob'
        LESS_THAN = 'lt'

    class FiltersMethod:
        OR = 'or'
        AND = 'and'

    class FiltersOrderByDirection:
        DESCENDING = 'descending'
        ASCENDING = 'ascending'

    class KnownFields:
        DIR = 'dir'
        ANNOTATED = 'annotated'
        FILENAME = 'filename'
        CREATED_AT = 'createdAt'
        UPDATED_AT = 'updatedAt'
        LABEL = 'label'
        NAME = 'name'
        HIDDEN = 'hidden'

    class ExecutionStatus:
        SUCCESS = 'success'
        FAILED = 'failed'
        IN_PROGRESS = 'inProgress'
        CREATED = 'created'

    class HttpMethod:
        GET = 'GET'
        POST = 'POST'
        DELETE = 'DELETE'
        PATCH = 'PATCH'

    class ViewAnnotationOptions:
        JSON = 'json'
        MASK = 'mask'
        INSTANCE = 'instance'

    class AnnotationFormat:
        COCO = 'coco'
        VOC = 'voc'
        YOLO = 'yolo'
        DATALOOP = 'dataloop'

    class InstanceCatalog:
        REGULAR_MICRO = 'regular-micro'
        REGULAR_XS = 'regular-xs'
        REGULAR_S = 'regular-s'
        REGULAR_M = 'regular-m'
        REGULAR_L = 'regular-l'
        REGULAR_XL = 'regular-xl'
        HIGHMEM_MICRO = 'highmem-micro'
        HIGHMEM_XS = 'highmem-xs'
        HIGHMEM_S = 'highmem-s'
        HIGHMEM_M = 'highmem-m'
        HIGHMEM_L = 'highmem-l'
        HIGHMEM_XL = 'highmem-xl'
        GPU_K80_S = 'gpu-k80-s'

    class LoggingLevel:
        DEBUG = 'debug'
        WARNING = 'warning'
        CRITICAL = 'critical'
        INFO = 'info'



import logging

from .. import entities, miscellaneous, exceptions, services

logger = logging.getLogger(name=__name__)


class Triggers:
    """
    Triggers repository
    """

    def __init__(self,
                 client_api: services.ApiClient,
                 project: entities.Project = None,
                 service: entities.Service = None,
                 project_id: str = None):
        self._client_api = client_api
        self._project = project
        self._service = service
        if project_id is None:
            if self._project is not None:
                project_id = self._project.id
            elif self._service is not None:
                project_id = self._service.project_id

        self._project_id = project_id

    ############
    # entities #
    ############
    @property
    def service(self) -> entities.Service:
        if self._service is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "service". need to set a Service entity or use service.triggers repository')
        assert isinstance(self._service, entities.Service)
        return self._service

    @service.setter
    def service(self, service: entities.Service):
        if not isinstance(service, entities.Service):
            raise ValueError('Must input a valid Service entity')
        self._service = service

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            if self._service is not None:
                self._project = self._service._project
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.triggers repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ###########
    # methods #
    ###########
    def create(self,
               service_id: str = None,
               webhook_id=None,
               name: str = None,
               filters=None,
               function_name=entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME,
               resource=entities.TriggerResource.ITEM,
               actions=None,
               active=True,
               execution_mode=entities.TriggerExecutionMode.ONCE,
               project_id=None,
               **kwargs) -> entities.Trigger:
        """
        Create a Trigger

        :param function_name:
        :param project_id:
        :param webhook_id:
        :param name:
        :param execution_mode:
        :param service_id: Id of services to be triggered
        :param filters: optional - Item/Annotation metadata filters, default = none
        :param resource: optional - Dataset/Item/Annotation/ItemStatus, default = Item
        :param actions: optional - Created/Updated/Deleted, default = create
        :param active: optional - True/False, default = True
        :return: Trigger entity
        """
        scope = kwargs.get('scope', None)

        if service_id is None and webhook_id is None:
            if self._service is not None:
                service_id = self._service.id

        # type
        if (service_id is None and webhook_id is None) or (service_id is not None and webhook_id is not None):
            raise exceptions.PlatformException('400', 'Must provide either service id or webhook id but not both')

        if name is None:
            if self._service is not None:
                name = self._service.name
            else:
                name = 'default_trigger'

        if filters is None:
            filters = dict()
        elif isinstance(filters, entities.Filters):
            filters = filters.prepare(query_only=True).get('filter', dict())

        if service_id is None:
            operation = {
                'type': 'webhook',
                'webhookId': webhook_id
            }
        else:
            operation = {
                'type': 'function',
                'serviceId': service_id,
                'functionName': function_name

            }

        if actions is not None:
            if not isinstance(actions, list):
                actions = [actions]
        else:
            actions = [entities.TriggerAction.CREATED]

        spec = {
            'filter': filters,
            'operation': operation,
            'resource': resource,
            'executionMode': execution_mode,
            'actions': actions
        }

        # payload
        if self._project_id is None and project_id is None:
            raise exceptions.PlatformException('400', 'Please provide a project id')
        elif project_id is None:
            project_id = self._project_id

        payload = {
            'active': active,
            'projectId': project_id,
            'name': name,
            'spec': spec
        }

        if scope is not None:
            logger.warning(
                "Only superuser is allowed to define a trigger's scope. "
                "If you are not a superuser you will not be able to perform this action")
            payload['scope'] = scope

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/triggers',
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Trigger.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          project=self._project,
                                          service=self._service)

    def get(self, trigger_id=None, trigger_name=None) -> entities.Trigger:
        """
        Get Trigger object

        :param trigger_name:
        :param trigger_id:
        :return: Trigger object
        """
        # request
        if trigger_id is not None:
            success, response = self._client_api.gen_request(
                req_type="get",
                path="/triggers/{}".format(trigger_id)
            )

            # exception handling
            if not success:
                raise exceptions.PlatformException(response)

            # return entity
            trigger = entities.Trigger.from_json(client_api=self._client_api,
                                                 _json=response.json(),
                                                 project=self._project,
                                                 service=self._service)
        else:
            if trigger_name is None:
                raise exceptions.PlatformException('400', 'Must provide either trigger name or trigger id')
            else:
                triggers = self.list(filters=entities.Filters(field='name', values=trigger_name,
                                                              resource=entities.FiltersResource.TRIGGER))
                if triggers.items_count == 0:
                    raise exceptions.PlatformException('404', 'Trigger not found')
                elif triggers.items_count == 1:
                    trigger = triggers[0]
                else:
                    raise exceptions.PlatformException('404',
                                                       'More than one trigger by name {} exist'.format(trigger_name))

        return trigger

    def delete(self, trigger_id=None, trigger_name=None):
        """
        Delete Trigger object

        :param trigger_name:
        :param trigger_id:
        :return: True
        """
        if trigger_id is None:
            if trigger_name is None:
                raise exceptions.PlatformException('400', 'Must provide either trigger name or trigger id')
            else:
                trigger_id = self.get(trigger_name=trigger_name).id
        # request
        success, response = self._client_api.gen_request(
            req_type="delete",
            path="/triggers/{}".format(trigger_id)
        )
        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        return True

    def update(self, trigger: entities.Trigger) -> entities.Trigger:
        """

        :param trigger: Trigger entity
        :return: Trigger entity
        """
        # payload
        payload = trigger.to_json()
        payload['spec'].pop('resource')
        payload['spec'].pop('actions')

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='/triggers/{}'.format(trigger.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Trigger.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          project=self._project,
                                          service=self._service)

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Trigger]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_trigger, trigger in enumerate(response_items):
            jobs[i_trigger] = pool.apply_async(entities.Trigger._protected_from_json,
                                               kwds={'client_api': self._client_api,
                                                     '_json': trigger,
                                                     'project': self._project,
                                                     'service': self._service})
        # wait for all jobs
        _ = [j.wait() for j in jobs]
        # get all results
        results = [j.get() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        triggers = miscellaneous.List([r[1] for r in results if r[0] is True])
        return triggers

    def _list(self, filters: entities.Filters):
        """
        List project triggers
        :return:
        """
        url = '/query/FaaS'

        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List project packages
        :return:
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.TRIGGER)
            if self._project is not None:
                filters.add(field='projectId', values=self._project.id)
            if self._service is not None:
                filters.add(field='spec.operation.serviceId', values=self._service.id)

        # assert type filters
        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException('400', 'Unknown filters type')

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def resource_information(self, resource, resource_type, action='Created'):
        """
        return which function should run on a item (based on global triggers)

        :param resource: 'Item' / 'Dataset' / etc
        :param resource_type: dictionary of the resource object
        :param action: 'Created' / 'Updated' / etc.

        """
        url = '/trigger-resource-information'

        payload = {'resource': resource_type,
                   'entity': resource.to_json(),
                   'action': action}
        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()



import logging
import math
import copy
import attr
from .. import services, miscellaneous

logger = logging.getLogger(name=__name__)


@attr.s
class PagedEntities:
    """
    Pages object
    """
    # api
    _client_api = attr.ib(type=services.ApiClient, repr=False)

    # params
    page_offset = attr.ib()
    page_size = attr.ib()
    filters = attr.ib()
    items_repository = attr.ib(repr=False)
    has_next_page = attr.ib(default=False)
    total_pages_count = attr.ib(default=0)
    items_count = attr.ib(default=0)

    # execution attribute
    _service_id = attr.ib(default=None, repr=False)
    _project_id = attr.ib(default=None, repr=False)
    _order_by_type = attr.ib(default=None, repr=False)
    _order_by_direction = attr.ib(default=None, repr=False)
    _execution_status = attr.ib(default=None, repr=False)
    _execution_resource_type = attr.ib(default=None, repr=False)
    _execution_resource_id = attr.ib(default=None, repr=False)
    _execution_function_name = attr.ib(default=None, repr=False)

    # items list
    items = attr.ib(default=miscellaneous.List(), repr=False)

    def process_result(self, result):
        if 'page_offset' in result:
            self.page_offset = result['page_offset']
        if 'page_size' in result:
            self.page_size = result['page_size']
        if 'hasNextPage' in result:
            self.has_next_page = result['hasNextPage']
        if 'totalItemsCount' in result:
            self.items_count = result['totalItemsCount']
        if 'totalPagesCount' in result:
            self.total_pages_count = result['totalPagesCount']
        if 'items' in result:
            items = self.items_repository._build_entities_from_response(response_items=result['items'])
        else:
            items = miscellaneous.List(list())
        return items

    def __getitem__(self, y):
        return self.items[y]

    def __len__(self):
        return self.items_count

    def __iter__(self):
        self.page_offset = 0
        self.has_next_page = True
        while self.has_next_page:
            self.get_page()
            self.page_offset += 1
            yield self.items

    def __reversed__(self):
        self.page_offset = self.total_pages_count - 1
        while True:
            self.get_page()
            yield self.items
            if self.page_offset == 0:
                break
            self.page_offset -= 1

    def return_page(self, page_offset=None, page_size=None):
        if page_size is None:
            page_size = self.page_size
        if page_offset is None:
            page_offset = self.page_offset

        if self.filters is not None:
            filters = copy.copy(self.filters)
            filters.page = page_offset
            filters.page_size = page_size
            result = self.items_repository._list(filters=filters)
            items = self.process_result(result)
            return items
        else:
            raise ValueError('Cant return page. Filters is empty')

    def get_page(self, page_offset=None, page_size=None):
        items = self.return_page(page_offset=page_offset,
                                 page_size=page_size)
        self.items = items

    def print(self):
        self.items.print()

    def next_page(self):
        """
        Brings the next page of items from host

        :return:
        """
        self.page_offset += 1
        self.get_page()

    def prev_page(self):
        """
        Brings the previous page of items from host

        :return:
        """
        self.page_offset -= 1
        self.get_page()

    def go_to_page(self, page=0):
        """
        Brings specified page of items from host

        :param page: page number
        :return:
        """
        self.page_offset = page
        self.get_page()

    def all(self):
        page_offset = 0
        page_size = 100
        total_pages = math.ceil(self.items_count / page_size)
        jobs = list()
        pool = self._client_api.thread_pools('item.page')
        while True:
            if page_offset <= total_pages:
                jobs.append(pool.apply_async(self.return_page, kwds={'page_offset': page_offset,
                                                                     'page_size': page_size}))
                page_offset += 1
            for i_job, job in enumerate(jobs):
                if job.ready():
                    for item in job.get():
                        yield item
                    jobs.remove(job)
            if len(jobs) == 0:
                break



import numpy as np
import traceback
import logging
import attr
import json
from PIL import Image

from .. import entities, PlatformException, repositories, ApiClient

logger = logging.getLogger(name=__name__)


class ViewAnnotationOptions:
    JSON = "json"
    MASK = "mask"
    INSTANCE = "instance"
    ANNOTATION_ON_IMAGE = "img_mask"
    VTT = "vtt"
    OBJECT_ID = "object_id"

    @staticmethod
    def list():
        return [ViewAnnotationOptions.JSON,
                ViewAnnotationOptions.MASK,
                ViewAnnotationOptions.INSTANCE,
                ViewAnnotationOptions.VTT,
                ViewAnnotationOptions.ANNOTATION_ON_IMAGE,
                ViewAnnotationOptions.OBJECT_ID]


@attr.s
class Annotation(entities.BaseEntity):
    """
    Annotations object
    """
    # annotation definition
    annotation_definition = attr.ib(repr=False, type=entities.BaseAnnotationDefinition)

    # platform
    id = attr.ib()
    url = attr.ib(repr=False)
    item_url = attr.ib(repr=False)
    item = attr.ib(repr=False)
    item_id = attr.ib()
    creator = attr.ib()
    createdAt = attr.ib()
    updatedBy = attr.ib(repr=False)
    updatedAt = attr.ib(repr=False)
    type = attr.ib()
    dataset_url = attr.ib(repr=False)

    # meta
    metadata = attr.ib(repr=False)
    fps = attr.ib(repr=False)
    hash = attr.ib(default=None, repr=False)
    dataset_id = attr.ib(default=None, repr=False)
    status = attr.ib(default=None, repr=False)
    object_id = attr.ib(default=None, repr=False)
    automated = attr.ib(default=None, repr=False)
    height = attr.ib(default=None)
    width = attr.ib(default=None)

    # snapshots
    frames = attr.ib(default=None, repr=False)
    current_frame = attr.ib(default=0, repr=False)

    # video attributes
    end_frame = attr.ib(default=0, repr=False)
    end_time = attr.ib(default=0, repr=False)
    start_frame = attr.ib(default=0)
    start_time = attr.ib(default=0)

    # temp
    platform_dict = attr.ib(default=None, repr=False)
    _dataset = attr.ib(repr=False, default=None)
    _datasets = attr.ib(repr=False, default=None)
    _annotations = attr.ib(repr=False, default=None)
    __client_api = attr.ib(default=None, repr=False)
    _items = attr.ib(repr=False, default=None)

    ####################################
    # annotation definition attributes #
    ####################################
    @property
    def parent_id(self):
        try:
            parent_id = self.metadata['system']['parentId']
        except KeyError:
            parent_id = None
        return parent_id

    @parent_id.setter
    def parent_id(self, parent_id):
        if 'system' not in self.metadata:
            self.metadata['system'] = dict()
        self.metadata['system']['parentId'] = parent_id

    @property
    def coordinates(self):
        color = None
        if self.type in ['binary']:
            color = self.color
        coordinates = self.annotation_definition.to_coordinates(color=color)
        return coordinates

    @property
    def _client_api(self) -> ApiClient:
        if self.__client_api is None:
            if self.item is None:
                raise PlatformException('400',
                                        'This action cannot be performed without an item entity. Please set item')
            else:
                self.__client_api = self.item._client_api
        assert isinstance(self.__client_api, ApiClient)
        return self.__client_api

    @property
    def dataset(self):
        if self._dataset is None:
            self._dataset = self.datasets.get(dataset_id=self.dataset_id)
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @property
    def annotations(self):
        if self._annotations is None:
            self._annotations = repositories.Annotations(client_api=self._client_api, item=self.item)
        assert isinstance(self._annotations, repositories.Annotations)
        return self._annotations

    @property
    def datasets(self):
        if self._datasets is None:
            self._datasets = repositories.Datasets(client_api=self._client_api)
        assert isinstance(self._datasets, repositories.Datasets)
        return self._datasets

    @property
    def items(self):
        if self._items is None:
            if self._datasets is not None:
                self._items = self._dataset.items
            elif self.item is not None:
                self._items = self.item.items
            else:
                self._items = repositories.Items(client_api=self._client_api, dataset=self._dataset)
        assert isinstance(self._items, repositories.Items)
        return self._items

    @property
    def x(self):
        return self.annotation_definition.x

    @property
    def y(self):
        return self.annotation_definition.y

    @property
    def rx(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.rx
        else:
            return None

    @property
    def ry(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.ry
        else:
            return None

    @property
    def angle(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.angle
        else:
            return None

    @property
    def geo(self):
        return self.annotation_definition.geo

    @geo.setter
    def geo(self, geo):
        self.annotation_definition.geo = geo

    @property
    def top(self):
        return self.annotation_definition.top

    @top.setter
    def top(self, top):
        self.annotation_definition.top = top

    @property
    def bottom(self):
        return self.annotation_definition.bottom

    @bottom.setter
    def bottom(self, bottom):
        self.annotation_definition.bottom = bottom

    @property
    def left(self):
        return self.annotation_definition.left

    @left.setter
    def left(self, left):
        self.annotation_definition.left = left

    @property
    def description(self):
        description = None
        if 'system' in self.metadata:
            description = self.metadata['system'].get('description', None)
        return description

    @description.setter
    def description(self, description):
        if 'system' in self.metadata:
            self.metadata['system']['description'] = description

    @property
    def right(self):
        return self.annotation_definition.right

    @right.setter
    def right(self, right):
        self.annotation_definition.right = right

    @property
    def last_frame(self):
        if len(self.frames) == 0:
            return 0
        return max(self.frames)

    @property
    def label(self):
        return self.annotation_definition.label

    @label.setter
    def label(self, label):
        self.annotation_definition.label = label

    @property
    def attributes(self):
        return self.annotation_definition.attributes

    @attributes.setter
    def attributes(self, attributes):
        if not isinstance(attributes, list):
            attributes = [attributes]
        self.annotation_definition.attributes = attributes

    @property
    def color(self):
        try:
            lower_keys = {key.lower(): label for key, label in self.dataset.labels_flat_dict.items()}
            if self.label.lower() in lower_keys:
                color = lower_keys[self.label.lower()].rgb
            else:
                color = None
            return color
        except Exception:
            return None

    ####################
    # frame attributes #
    ####################
    @property
    def frame_num(self):
        if len(self.frames) > 0:
            return self.current_frame
        else:
            return self.start_frame

    @frame_num.setter
    def frame_num(self, frame_num):
        if frame_num != self.current_frame:
            self.frames[self.current_frame].frame_num = frame_num
            self.frames[frame_num] = self.frames[self.current_frame]
            self.frames.pop(self.current_frame)

    @property
    def fixed(self):
        if len(self.frames) > 0:
            return self.frames[self.current_frame].fixed
        else:
            return False

    @fixed.setter
    def fixed(self, fixed):
        if len(self.frames) > 0:
            self.frames[self.current_frame].fixed = fixed

    @property
    def is_video(self):
        if len(self.frames) == 0:
            return False
        else:
            return True

    ################
    # polygon only #
    ################
    @property
    def is_open(self):
        is_open = None
        if self.type in ['segment', 'polyline']:
            is_open = self.annotation_definition.is_open
        return is_open

    @is_open.setter
    def is_open(self, is_open):
        if self.type in ['segment']:
            self.annotation_definition.is_open = is_open
        else:
            logger.warning('type {} annotation does not have attribute is_open'.format(self.type))

    ##################
    # entity methods #
    ##################
    def update_status(self, status):
        """
        Open an issue on the annotation
        """
        return self.annotations.update_status(annotation=self, status=status)

    def delete(self):
        """
        Remove an annotation from item
        :return: True
        """
        return self.annotations.delete(annotation_id=self.id)

    def update(self, system_metadata=False):
        """
        Update an existing annotation in host.
        :param system_metadata:
        :return: Annotation object
        """
        return self.annotations.update(annotations=self,
                                       system_metadata=system_metadata)[0]

    def upload(self):
        """
        Create a new annotation in host
        :return:
        """
        return self.annotations.upload(annotations=self)[0]

    def download(self,
                 filepath,
                 annotation_format=ViewAnnotationOptions.MASK,
                 height=None,
                 width=None,
                 thickness=1,
                 with_text=False):
        """
        Save annotation to file
        :param filepath: local path to where annotation will be downloaded to
        :param annotation_format: options: dl.ViewAnnotationOptions.list()
        :param height: image height
        :param width: image width
        :param thickness: thickness
        :param with_text: get mask with text
        :return: filepath
        """
        if annotation_format == ViewAnnotationOptions.JSON:
            with open(filepath, 'w') as f:
                json.dump(self.to_json(), f, indent=2)
        else:
            mask = self.show(thickness=thickness,
                             with_text=with_text,
                             height=height,
                             width=width,
                             annotation_format=annotation_format)
            img = Image.fromarray(mask.astype(np.uint8))
            img.save(filepath)
        return filepath

    def set_frame(self, frame):
        """
        Set annotation to frame state
        :param frame: frame number
        :return: True
        """
        if frame in self.frames:
            self.current_frame = frame
            self.annotation_definition = self.frames[frame].annotation_definition
            return True
        else:
            return False

    ############
    # Plotting #
    ############
    def show(self, image=None, thickness=None, with_text=False, height=None, width=None,
             annotation_format=ViewAnnotationOptions.MASK, color=None):
        """
        Show annotations

        :param image: empty or image to draw on
        :param height: height
        :param width: width
        :param thickness: line thickness
        :param with_text: add label to annotation
        :param annotation_format: dl.ViewAnnotationOptions.list()
        :param color: optional
        :return: ndarray of the annotations
        """
        try:
            import cv2
        except ImportError:
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        # height/width
        if height is None:
            if self.item is None or self.item.height is None:
                raise PlatformException('400', 'must provide item width and height')
            height = self.item.height
        if width is None:
            if self.item is None or self.item.width is None:
                raise PlatformException('400', 'must provide item width and height')
            width = self.item.width

        # color
        if color is None:
            if self.color is not None:
                color = self.color
            else:
                if self.label.lower() not in ['approved', 'completed']:
                    logger.warning('No color given for label: {}, random color will be selected'.format(self.label))
                color = (127, 127, 127)
        if (isinstance(color, list) or isinstance(color, tuple)) and len(color) == 3:
            # if color is a list or tuple and size of 3 - add alpha
            color = (color[0], color[1], color[2], 255)
            if image is not None:
                # check matching dimensions
                if len(image.shape) == 2:
                    # image is gray
                    image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGBA)
                elif image.shape[2] == 3:
                    image = cv2.cvtColor(image, cv2.COLOR_RGB2RGBA)
                elif image.shape[2] == 4:
                    pass
                else:
                    raise PlatformException(error='400',
                                            message='Image size doesnt match colors. img shape: {}, color: {}'.format(
                                                image.shape,
                                                color))
        # show annotation
        if image is None:
            image = np.zeros((height, width, len(color)), dtype=np.uint8)
            if image.shape[2] == 1:
                image = np.squeeze(image)
        return self.annotation_definition.show(image=image,
                                               thickness=thickness,
                                               with_text=with_text,
                                               height=height,
                                               width=width,
                                               annotation_format=annotation_format,
                                               color=color)

    #######
    # I/O #
    #######
    @classmethod
    def new(cls,
            item=None,
            annotation_definition=None,
            object_id=None,
            automated=None,
            metadata=None,
            frame_num=None,
            parent_id=None,
            start_time=None,
            height=None,
            width=None):
        """
        Create a new annotation object annotations

        :param start_time:
        :param width: annotation item's width
        :param height: annotation item's height
        :param item: item to annotate
        :param annotation_definition: annotation type object
        :param object_id: object_id
        :param automated: is automated
        :param metadata: metadata
        :param frame_num: optional - first frame number if video annotation
        :param parent_id: add parent annotation ID
        :return: annotation object
        """
        if frame_num is None:
            frame_num = 0

        # init annotations
        if metadata is None:
            metadata = dict()

        # add parent
        if parent_id is not None:
            if 'system' not in metadata:
                metadata['system'] = dict()
            metadata['system']['parentId'] = parent_id

        # frames
        frames = dict()

        # handle fps
        if item is not None and item.fps is not None:
            fps = item.fps
        else:
            fps = None

        # get type
        ann_type = None
        if annotation_definition is not None:
            ann_type = annotation_definition.type

        # dataset
        dataset_url = None
        dataset_id = None
        if item is not None:
            dataset_url = item.dataset_url
            dataset_id = item.datasetId

        if start_time is None:
            if fps is not None and frame_num is not None:
                start_time = frame_num / fps if fps != 0 else 0
            else:
                start_time = 0

        if frame_num is None:
            frame_num = 0

        return cls(
            # annotation_definition
            annotation_definition=annotation_definition,

            # platform
            id=None,
            url=None,
            item_url=None,
            item=item,
            item_id=None,
            creator=None,
            createdAt=None,
            updatedBy=None,
            updatedAt=None,
            object_id=object_id,
            type=ann_type,
            dataset_url=dataset_url,
            dataset_id=dataset_id,
            height=height,
            width=width,

            # meta
            metadata=metadata,
            fps=fps,
            status=None,
            automated=automated,

            # snapshots
            frames=frames,

            # video only attributes
            end_frame=frame_num,
            end_time=0,
            start_frame=frame_num,
            start_time=start_time,

            # temp
            platform_dict=dict(),
        )

    def add_frames(self, annotation_definition, frame_num=None, end_frame_num=None, start_time=None, end_time=None,
                   fixed=True):
        # handle fps
        if self.fps is None:
            if self.item is not None:
                if self.item.fps is not None:
                    self.fps = self.item.fps
        if self.fps is None:
            raise PlatformException('400', 'Annotation must have fps in order to perform this action')

        if frame_num is None:
            frame_num = int(np.round(start_time * self.fps))

        if end_frame_num is None:
            if end_time is not None:
                end_frame_num = int(np.round(end_time * self.fps))
            else:
                end_frame_num = frame_num

        for frame in range(frame_num, end_frame_num + 1):
            self.add_frame(annotation_definition=annotation_definition,
                           frame_num=frame,
                           fixed=fixed)

    def add_frame(self, annotation_definition, frame_num=None, fixed=True):
        """
        Add a frame state to annotation

        :param annotation_definition: annotation type object - must be same type as annotation
        :param fixed: is fixed
        :param frame_num: frame number
        :return: annotation object
        """
        # handle fps
        if self.fps is None:
            if self.item is not None:
                if self.item.fps is not None:
                    self.fps = self.item.fps
        if self.fps is None:
            raise PlatformException('400', 'Annotation must have fps in order to perform this action')

        # if this is first frame
        if self.annotation_definition is None:

            if frame_num is None:
                frame_num = 0
            self.start_frame = frame_num
            self.current_frame = frame_num
            self.end_frame = frame_num
            self.start_time = frame_num / self.fps if self.fps != 0 else 0

            frame = FrameAnnotation.new(annotation_definition=annotation_definition,
                                        frame_num=frame_num,
                                        fixed=fixed,
                                        annotation=self)

            self.frames[frame_num] = frame
            self.set_frame(frame_num)
            self.end_time = self.end_frame / self.fps if self.fps != 0 else 0
            self.type = annotation_definition.type

            return True

        # check if type matches annotation
        if not isinstance(annotation_definition, type(self.annotation_definition)):
            raise PlatformException('400', 'All frames in annotation must have same type')

        # find frame number
        if frame_num is None:
            frame_num = self.last_frame + 1
        elif frame_num < self.start_frame:
            self.start_frame = frame_num

        # add frame to annotation
        if not self.is_video:
            # create first frame from annotation definition
            frame = FrameAnnotation.new(annotation_definition=self.annotation_definition,
                                        frame_num=self.last_frame,
                                        fixed=fixed,
                                        annotation=self)

            self.frames[self.start_frame] = frame

        # create new time annotations
        frame = FrameAnnotation.new(annotation_definition=annotation_definition,
                                    frame_num=frame_num,
                                    fixed=fixed,
                                    annotation=self)

        self.frames[frame_num] = frame
        self.end_frame = max(self.frames)
        self.end_time = self.end_frame / self.fps

        return True

    @staticmethod
    def _protected_from_json(_json,
                             item=None,
                             client_api=None,
                             annotations=None,
                             is_video=None,
                             fps=None,
                             item_metadata=None):
        """
        Same as from_json but with try-except to catch if error
        :param client_api:
        :param annotations:
        :param is_video:
        :param fps:
        :param item_metadata:
        :param _json: platform json
        :param item: item
        :return: annotation object
        """
        try:
            annotation = Annotation.from_json(_json=_json,
                                              item=item,
                                              client_api=client_api,
                                              annotations=annotations,
                                              is_video=is_video,
                                              fps=fps,
                                              item_metadata=item_metadata)
            status = True
        except Exception:
            annotation = traceback.format_exc()
            status = False
        return status, annotation

    @classmethod
    def from_json(cls,
                  _json,
                  item=None,
                  client_api=None,
                  annotations=None,
                  is_video=None,
                  fps=None,
                  item_metadata=None):
        """
        Create an annotation object from platform json
        :param client_api:
        :param annotations:
        :param is_video:
        :param fps:
        :param item_metadata:
        :param _json: platform json
        :param item: item
        :return: annotation object
        """
        if item_metadata is None:
            item_metadata = dict()

        if is_video is None:
            if item is None:
                is_video = False
            else:
                # get item type
                if 'video' in item.mimetype:
                    is_video = True

        item_url = _json.get('item', item.url if item is not None else None)
        item_id = _json.get('itemId', item.id if item is not None else None)
        dataset_url = _json.get('dataset', item.dataset_url if item is not None else None)
        dataset_id = _json.get('datasetId', item.datasetId if item is not None else None)

        # get id
        if 'id' in _json:
            annotation_id = _json['id']
        elif '_id' in _json:
            annotation_id = _json['_id']
        else:
            raise PlatformException('400', 'missing id in annotation json')

        # get metadata, status, attributes and object id
        object_id = None
        status = None
        attributes = _json.get('attributes', list())
        metadata = _json.get('metadata', dict())
        if 'system' in metadata:
            object_id = _json['metadata']['system'].get('objectId', object_id)
            status = _json['metadata']['system'].get('status', status)

        first_frame_attributes = list()
        first_frame_coordinates = list()
        first_frame_number = 0
        first_frame_start_time = 0
        automated = None
        end_frame = None
        start_time = 0
        start_frame = 0

        ############
        # if video #
        ############
        if is_video:
            # get fps
            if item is not None and item.fps is not None:
                fps = item.fps
            if fps is None:
                fps = item_metadata.get('fps', 25)

            # get video-only attributes
            end_time = 1.5
            # get first frame attribute
            first_frame_attributes = _json.get('attributes', first_frame_attributes)
            # get first frame coordinates
            first_frame_coordinates = _json.get('coordinates', first_frame_coordinates)
            if 'system' in metadata:
                # get first frame number
                first_frame_number = _json['metadata']['system'].get('frame', first_frame_number)
                # get first frame start time
                start_time = _json['metadata']['system'].get('startTime', first_frame_start_time)
                # get first frame number
                start_frame = _json['metadata']['system'].get('frame', start_frame)
                automated = _json['metadata']['system'].get('automated', automated)
                end_frame = _json['metadata']['system'].get('endFrame', end_frame)
                end_time = _json['metadata']['system'].get('endTime', end_time)

            annotation_definition = None
        ################
        # if not video #
        ################
        else:
            # get coordinates
            coordinates = _json.get('coordinates', list())
            # set video only attributes
            end_time = 0
            # get automated
            if 'system' in metadata:
                automated = metadata['system'].get('automated', automated)
            # set annotation definition
            def_dict = {'type': _json['type'],
                        'coordinates': coordinates,
                        'label': _json['label'],
                        'attributes': attributes}
            if _json['type'] == 'segment':
                is_open = False
                if 'system' in metadata:
                    is_open = metadata['system'].get('isOpen', is_open)
                def_dict['is_open'] = is_open
            annotation_definition = FrameAnnotation.json_to_annotation_definition(def_dict)

        frames = dict()

        # init annotation
        annotation = cls(
            # temp
            platform_dict=_json,
            # annotation definition
            annotation_definition=annotation_definition,
            # platform
            id=annotation_id,
            url=_json.get('url', None),
            item_url=item_url,
            item=item,
            item_id=item_id,
            dataset_url=dataset_url,
            dataset_id=dataset_id,
            creator=_json['creator'],
            createdAt=_json['createdAt'],
            updatedBy=_json['updatedBy'],
            updatedAt=_json['updatedAt'],
            hash=_json.get('hash', None),
            object_id=object_id,
            type=_json['type'],
            width=item_metadata.get('width', None),
            height=item_metadata.get('height', None),
            # meta
            metadata=metadata,
            fps=fps,
            status=status,
            # snapshots
            frames=frames,
            # video attributes
            automated=automated,
            end_frame=end_frame,
            end_time=end_time,
            start_frame=start_frame,
            annotations=annotations,
            start_time=start_time
        )
        annotation.__client_api = client_api

        #################
        # if has frames #
        #################
        if is_video:
            if annotation.type == 'class' or annotation.type == 'subtitle':
                if end_frame is None:
                    end_frame = start_frame
                # for class type annotation create frames
                # make copies of the head annotations for all frames in it
                for frame_num in range(start_frame, end_frame + 1):
                    snapshot = {
                        'frame': frame_num,
                        'attributes': first_frame_attributes,
                        'coordinates': first_frame_coordinates,
                        'fixed': True,
                        'label': _json['label'],
                        'type': annotation.type,
                    }
                    frame = FrameAnnotation.from_snapshot(_json=snapshot,
                                                          annotation=annotation,
                                                          fps=fps)
                    annotation.frames[frame.frame_num] = frame
            else:
                # set first frame
                snapshot = {
                    'attributes': first_frame_attributes,
                    'coordinates': first_frame_coordinates,
                    'fixed': True,
                    'frame': first_frame_number,
                    'label': _json['label'],
                    'type': annotation.type,
                }

                # add first frame
                frame = FrameAnnotation.from_snapshot(_json=snapshot,
                                                      annotation=annotation,
                                                      fps=fps)
                annotation.frames[frame.frame_num] = frame
                annotation.annotation_definition = frame.annotation_definition
                # put snapshots if there are any
                for snapshot in _json['metadata']['system']['snapshots_']:
                    frame = FrameAnnotation.from_snapshot(_json=snapshot,
                                                          annotation=annotation,
                                                          fps=fps)
                    annotation.frames[frame.frame_num] = frame

            annotation.annotation_definition = annotation.frames[min(frames)].annotation_definition

        return annotation

    def to_json(self):
        """
        Convert annotation object to a platform json representation
        :return: platform json
        """
        if len(self.frames) > 0:
            self.set_frame(min(self.frames))
        _json = attr.asdict(self,
                            filter=attr.filters.include(attr.fields(Annotation).id,
                                                        attr.fields(Annotation).url,
                                                        attr.fields(Annotation).metadata,
                                                        attr.fields(Annotation).creator,
                                                        attr.fields(Annotation).hash,
                                                        attr.fields(Annotation).createdAt,
                                                        attr.fields(Annotation).updatedBy,
                                                        attr.fields(Annotation).updatedAt,
                                                        attr.fields(Annotation).metadata,
                                                        attr.fields(Annotation).createdAt,
                                                        attr.fields(Annotation).updatedBy))

        # property attributes                                                
        _json['itemId'] = self.item_id
        _json['item'] = self.item_url
        _json['label'] = self.label
        _json['attributes'] = self.attributes
        _json['dataset'] = self.dataset_url

        if self.item is not None and self.dataset_id is None:
            _json['datasetId'] = self.item.datasetId
        else:
            _json['datasetId'] = self.dataset_id

        # polyline to segment
        if self.type == 'polyline':
            _json['type'] = 'segment'
        else:
            _json['type'] = self.type
        if self.type == 'polyline':
            _json['coordinates'] = [self.coordinates]
        else:
            _json['coordinates'] = self.coordinates

        # add system metadata
        if 'system' not in _json['metadata']:
            _json['metadata']['system'] = dict()
        if self.automated is not None:
            _json['metadata']['system']['automated'] = self.automated
        if self.object_id is not None:
            _json['metadata']['system']['objectId'] = self.object_id
        if self.status is not None:
            _json['metadata']['system']['status'] = self.status
        if self.type in ['segment', 'polyline']:
            _json['metadata']['system']['isOpen'] = self.is_open

        # add frame info
        if self.is_video:
            # get all snapshots but the first one
            snapshots = list()
            first_frame_num = min(self.frames)
            for frame_num in sorted(self.frames):
                if frame_num == first_frame_num:
                    # ignore first frame in snapshots
                    continue
                snapshots.append(self.frames[frame_num].to_snapshot())
            # add metadata to json
            _json['metadata']['system']['frame'] = self.current_frame
            _json['metadata']['system']['startTime'] = self.start_time
            _json['metadata']['system']['endTime'] = self.end_time
            if self.end_frame is not None:
                _json['metadata']['system']['endFrame'] = self.end_frame

            # add snapshots only if classification
            if self.type not in ['class', 'subtitle']:
                _json['metadata']['system']['snapshots_'] = snapshots
        else:
            # remove metadata if empty
            if len(_json['metadata']['system']) == 0:
                _json['metadata'].pop('system')
                if len(_json['metadata']) == 0:
                    _json.pop('metadata')

        return _json


@attr.s
class FrameAnnotation(entities.BaseEntity):
    """
    FrameAnnotation object
    """
    # parent annotation
    annotation = attr.ib()

    # annotations
    annotation_definition = attr.ib()

    # multi
    frame_num = attr.ib()
    fixed = attr.ib()

    ################################
    # parent annotation attributes #
    ################################

    @property
    def status(self):
        return self.annotation.status

    @property
    def timestamp(self):
        if self.annotation.fps is not None and self.frame_num is not None:
            return self.frame_num / self.annotation.fps if self.annotation.fps != 0 else None

    ####################################
    # annotation definition attributes #
    ####################################
    @property
    def type(self):
        return self.annotation.type

    @property
    def label(self):
        return self.annotation_definition.label

    @property
    def attributes(self):
        return self.annotation_definition.attributes

    @property
    def geo(self):
        return self.annotation_definition.geo

    @property
    def top(self):
        return self.annotation_definition.top

    @property
    def bottom(self):
        return self.annotation_definition.bottom

    @property
    def left(self):
        return self.annotation_definition.left

    @property
    def right(self):
        return self.annotation_definition.right

    @property
    def color(self):
        if self.annotation.item is None:
            return 255, 255, 255
        else:
            label = None
            for label in self.annotation.item.dataset.labels:
                if label.tag.lower() == self.label.lower():
                    return label.rgb
            if label is None:
                return 255, 255, 255

    @property
    def coordinates(self):
        coordinates = self.annotation_definition.to_coordinates(color=self.color)
        return coordinates

    @property
    def x(self):
        return self.annotation_definition.x

    @property
    def y(self):
        return self.annotation_definition.y

    @property
    def rx(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.rx
        else:
            return None

    @property
    def ry(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.ry
        else:
            return None

    @property
    def angle(self):
        if self.annotation_definition.type == 'ellipse':
            return self.annotation_definition.angle
        else:
            return None

    ######################
    # annotation methods #
    ######################
    def show(self, **kwargs):
        """
        Show annotation as ndarray
        :param kwargs: see annotation definition
        :return: ndarray of the annotation
        """
        return self.annotation_definition.show(**kwargs)

    @staticmethod
    def json_to_annotation_definition(_json):
        if _json['type'] == 'segment':
            annotation = entities.Polygon.from_json(_json)
        elif _json['type'] == 'polyline':
            annotation = entities.Polyline.from_json(_json)
        elif _json['type'] == 'box':
            annotation = entities.Box.from_json(_json)
        elif _json['type'] == 'point':
            annotation = entities.Point.from_json(_json)
        elif _json['type'] == 'binary':
            annotation = entities.Segmentation.from_json(_json)
        elif _json['type'] == 'class':
            annotation = entities.Classification.from_json(_json)
        elif _json['type'] == 'subtitle':
            annotation = entities.Subtitle.from_json(_json)
        elif _json['type'] == 'ellipse':
            annotation = entities.Ellipse.from_json(_json)
        elif _json['type'] == 'comparison':
            annotation = entities.Comparison.from_json(_json)
        elif _json['type'] == 'note':
            annotation = entities.Note.from_json(_json)
        else:
            annotation = entities.UndefinedAnnotationType.from_json(_json)
        return annotation

    #######
    # I/O #
    #######
    @classmethod
    def new(cls, annotation, annotation_definition, frame_num, fixed):
        return cls(
            # annotations
            annotation=annotation,
            annotation_definition=annotation_definition,

            # multi
            frame_num=frame_num,
            fixed=fixed
        )

    @classmethod
    def from_snapshot(cls, annotation, _json, fps):
        # get annotation class
        _json['type'] = annotation.type
        annotation_definition = cls.json_to_annotation_definition(_json=_json)

        frame_num = _json.get('frame', annotation.last_frame + 1)

        return cls(
            # annotations
            annotation=annotation,
            annotation_definition=annotation_definition,

            # multi
            frame_num=frame_num,
            fixed=_json.get('fixed', False)
        )

    def to_snapshot(self):
        snapshot_dict = {'frame': self.frame_num,
                         'fixed': self.fixed,
                         'label': self.label,
                         'attributes': self.attributes,
                         'type': self.type,
                         'data': self.coordinates}
        return snapshot_dict



import attr
import traceback
from collections import namedtuple

from .. import entities, services, exceptions, repositories


class TriggerResource:
    ITEM = "Item"
    DATASET = "Dataset"
    ANNOTATION = "Annotation"
    ITEM_STATUS = "ItemStatus"


class TriggerAction:
    CREATED = "Created"
    UPDATED = "Updated"
    DELETED = "Deleted"


class TriggerExecutionMode:
    ONCE = "Once"
    ALWAYS = "Always"


@attr.s
class Trigger(entities.BaseEntity):
    """
    Trigger Entity
    """
    #######################
    # Platform attributes #
    #######################
    id = attr.ib()
    url = attr.ib(repr=False)
    createdAt = attr.ib()
    updatedAt = attr.ib(repr=False)
    creator = attr.ib()
    name = attr.ib()
    active = attr.ib()
    scope = attr.ib()

    ###################
    # spec attributes #
    ###################
    _resource = attr.ib()
    _actions = attr.ib()
    execution_mode = attr.ib(repr=False)
    # name change
    function_name = attr.ib()
    service_id = attr.ib()
    webhook_id = attr.ib()

    ########
    # temp #
    ########
    special = attr.ib(repr=False)

    ##############################
    # different name in platform #
    ##############################
    filters = attr.ib()
    project_id = attr.ib()
    _spec = attr.ib()

    ##################
    # SDK attributes #
    ##################
    _service = attr.ib(repr=False)
    _project = attr.ib(repr=False)
    _client_api = attr.ib(type=services.ApiClient, repr=False)
    _op_type = attr.ib(default='service')
    _repositories = attr.ib(repr=False)

    @staticmethod
    def _protected_from_json(_json, client_api, project, service=None):
        """
        Same as from_json but with try-except to catch if error
        :param _json:
        :param client_api:
        :param project:
        :return:
        """
        try:
            trigger = Trigger.from_json(_json=_json,
                                        client_api=client_api,
                                        project=project,
                                        service=service)
            status = True
        except Exception:
            trigger = traceback.format_exc()
            status = False
        return status, trigger

    # noinspection PyShadowingBuiltins
    @classmethod
    def from_json(cls, _json, client_api, project, service=None):
        spec = _json.get('spec', dict())
        operation = spec.get('operation', dict())
        filters = spec.get('filter', dict())
        op_type = operation.get('type', None)

        if op_type == 'function':
            service_id = operation.get('serviceId', None)
            webhook_id = None
        elif op_type == 'webhook':
            webhook_id = operation.get('webhookId', None)
            service_id = None
        else:
            raise exceptions.PlatformException('400', 'unknown trigger operation type: {}'.format(op_type))

        project_id = _json.get('projectId', None)
        if project_id is not None and project is not None:
            if project_id != project.id:
                project = None

        return cls(
            execution_mode=spec.get('executionMode', None),
            updatedAt=_json.get('updatedAt', None),
            createdAt=_json.get('createdAt', None),
            resource=spec.get('resource', None),
            creator=_json.get('creator', None),
            special=_json.get('special', None),
            actions=spec.get('actions', None),
            active=_json.get('active', None),
            function_name=operation.get('functionName', None),
            scope=_json.get('scope', None),
            name=_json.get('name', None),
            service_id=service_id,
            project_id=project_id,
            url=_json.get('url', None),
            webhook_id=webhook_id,
            client_api=client_api,
            filters=filters,
            project=project,
            service=service,
            id=_json['id'],
            op_type=op_type,
            spec=spec,
        )

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['services'])
        if self._project is None:
            services = repositories.Services(client_api=self._client_api, project=self._project)
        else:
            services = self._project.services

        r = reps(services=services)
        return r

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    ############
    # entities #
    ############

    @property
    def project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id,
                                                                                   fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def service(self):
        if self._service is None:
            self._service = self.services.get(service_id=self.service_id, fetch=None)
        assert isinstance(self._service, entities.Service)
        return self._service

    @property
    def resource(self):
        return self._resource

    @resource.setter
    def resource(self, value):
        raise exceptions.PlatformException('400', 'Trigger resource cannot be changed! Please create a new trigger')

    @property
    def actions(self):
        return self._actions

    @actions.setter
    def actions(self, value):
        raise exceptions.PlatformException('400', 'Trigger resource cannot be actions! Please create a new trigger')

    ###########
    # methods #
    ###########
    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        # get excluded
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Trigger)._client_api,
                                                              attr.fields(Trigger).project_id,
                                                              attr.fields(Trigger)._project,
                                                              attr.fields(Trigger)._service,
                                                              attr.fields(Trigger).special,
                                                              attr.fields(Trigger).filters,
                                                              attr.fields(Trigger)._op_type,
                                                              attr.fields(Trigger)._spec,
                                                              attr.fields(Trigger)._resource,
                                                              attr.fields(Trigger)._repositories,
                                                              attr.fields(Trigger)._actions,
                                                              attr.fields(Trigger).service_id,
                                                              attr.fields(Trigger).webhook_id,
                                                              attr.fields(Trigger).execution_mode,
                                                              attr.fields(Trigger).function_name
                                                              ))

        # rename
        _json['projectId'] = self.project_id
        operation = {
            'type': self._op_type,
            'functionName': self.function_name
        }

        if self._op_type == 'function':
            operation['serviceId'] = self.service_id
        elif self._op_type == 'webhook':
            operation['webhookId'] = self.webhook_id

        _json['spec'] = {
            'filter': self.filters,
            'executionMode': self.execution_mode,
            'resource': self.resource,
            'actions': self.actions,
            'operation': operation,
        }

        return _json

    def delete(self):
        """
        Delete Trigger object

        :return: True
        """
        return self.project.triggers.delete(trigger_id=self.id)

    def update(self):
        """

        :return: Trigger entity
        """
        return self.project.triggers.update(trigger=self)



from collections import namedtuple
import traceback
import logging
import attr
import copy
import os

from .. import repositories, entities, services, exceptions

logger = logging.getLogger(name=__name__)


class ItemStatus:
    COMPLETED = "completed"
    APPROVED = "approved"
    DISCARDED = "discarded"


@attr.s
class Item(entities.BaseEntity):
    """
    Item object
    """
    # item information
    annotations_link = attr.ib(repr=False)
    dataset_url = attr.ib()
    thumbnail = attr.ib(repr=False)
    createdAt = attr.ib()
    datasetId = attr.ib()
    annotated = attr.ib(repr=False)
    metadata = attr.ib(repr=False)
    filename = attr.ib()
    stream = attr.ib(repr=False)
    name = attr.ib()
    type = attr.ib()
    url = attr.ib(repr=False)
    id = attr.ib()
    hidden = attr.ib(repr=False)
    dir = attr.ib(repr=False)

    # name change
    annotations_count = attr.ib()

    # api
    _client_api = attr.ib(type=services.ApiClient, repr=False)
    _platform_dict = attr.ib(repr=False)

    # entities
    _dataset = attr.ib(repr=False)
    _project = attr.ib(repr=False)

    # repositories
    _repositories = attr.ib(repr=False)

    @staticmethod
    def _protected_from_json(_json, client_api, dataset=None):
        """
        Same as from_json but with try-except to catch if error
        :param _json:
        :param client_api:
        :param dataset:
        :return:
        """
        try:
            item = Item.from_json(_json=_json,
                                  client_api=client_api,
                                  dataset=dataset)
            status = True
        except Exception:
            item = traceback.format_exc()
            status = False
        return status, item

    @classmethod
    def from_json(cls, _json, client_api, dataset=None, project=None, is_fetched=True):
        """
        Build an item entity object from a json
        :param project:
        :param _json: _json response from host
        :param dataset: dataset in which the annotation's item is located
        :param client_api: client_api
        :param is_fetched: is Entity fetched from Platform
        :return: Item object
        """
        metadata = _json.get('metadata', dict())
        inst = cls(
            # sdk
            platform_dict=copy.deepcopy(_json),
            client_api=client_api,
            dataset=dataset,
            project=project,
            # params
            annotations_link=_json.get('annotations', None),
            thumbnail=_json.get('thumbnail', None),
            datasetId=_json.get('datasetId', None),
            annotated=_json.get('annotated', None),
            dataset_url=_json.get('dataset', None),
            createdAt=_json.get('createdAt', None),
            annotations_count=_json.get('annotationsCount', None),
            hidden=_json.get('hidden', False),
            stream=_json.get('stream', None),
            dir=_json.get('dir', None),
            filename=_json.get('filename', None),
            metadata=metadata,
            name=_json.get('name', None),
            type=_json.get('type', None),
            url=_json.get('url', None),
            id=_json.get('id', None))
        inst.is_fetched = is_fetched
        return inst

    ############
    # entities #
    ############

    @property
    def dataset(self):
        if self._dataset is None:
            if self._dataset is None:
                self._dataset = self.datasets.get(dataset_id=self.datasetId, fetch=None)
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @property
    def project(self):
        if self._project is None:
            if self._dataset is None:
                if self._dataset is None:
                    self._dataset = self.datasets.get(dataset_id=self.datasetId, fetch=None)
            self._project = self._dataset.project
            if self._project is None:
                raise exceptions.PlatformException(error='2001',
                                                   message='Missing entity "project".')
        assert isinstance(self._project, entities.Project)
        return self._project

    ################
    # repositories #
    ################

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['annotations', 'datasets', 'items', 'codebases', 'artifacts', 'modalities'])
        reps.__new__.__defaults__ = (None, None, None, None, None, None)

        if self._dataset is None:
            items = repositories.Items(client_api=self._client_api,
                                       dataset=self._dataset,
                                       dataset_id=self.datasetId,
                                       datasets=repositories.Datasets(client_api=self._client_api, project=None))
            datasets = items.datasets
        else:
            items = self.dataset.items
            datasets = self.dataset.datasets

        r = reps(annotations=repositories.Annotations(client_api=self._client_api,
                                                      dataset_id=self.datasetId,
                                                      item=self,
                                                      dataset=self._dataset),
                 items=items,
                 datasets=datasets,
                 codebases=None,
                 artifacts=None,
                 modalities=Modalities(item=self))
        return r

    @property
    def modalities(self):
        assert isinstance(self._repositories.modalities, Modalities)
        return self._repositories.modalities

    @property
    def annotations(self):
        assert isinstance(self._repositories.annotations, repositories.Annotations)
        return self._repositories.annotations

    @property
    def datasets(self):
        assert isinstance(self._repositories.datasets, repositories.Datasets)
        return self._repositories.datasets

    @property
    def items(self):
        assert isinstance(self._repositories.items, repositories.Items)
        if self._repositories.items._dataset is None:
            self._repositories.items._dataset = self.dataset
        return self._repositories.items

    ##############
    # Properties #
    ##############
    @property
    def height(self):
        return self.metadata.get('system', dict()).get('height', None)

    @height.setter
    def height(self, val):
        if 'system' not in self.metadata:
            self.metadata['system'] = dict()
        self.metadata['system']['height'] = val

    @property
    def width(self):
        return self.metadata.get('system', dict()).get('width', None)

    @width.setter
    def width(self, val):
        if 'system' not in self.metadata:
            self.metadata['system'] = dict()
        self.metadata['system']['width'] = val

    @property
    def fps(self):
        return self.metadata.get('fps', None)

    @fps.setter
    def fps(self, val):
        self.metadata['fps'] = val

    @property
    def mimetype(self):
        return self.metadata.get('system', dict()).get('mimetype', None)

    @property
    def size(self):
        return self.metadata.get('system', dict()).get('size', None)

    @property
    def system(self):
        return self.metadata.get('system', dict())

    ###########
    # Functions #
    ###########
    def to_json(self):
        """
        Returns platform _json format of object
        :return: platform json format of object
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(Item)._repositories,
                                                        attr.fields(Item)._dataset,
                                                        attr.fields(Item)._project,
                                                        attr.fields(Item)._client_api,
                                                        attr.fields(Item)._platform_dict,
                                                        attr.fields(Item).annotations_count,
                                                        attr.fields(Item).dataset_url,
                                                        attr.fields(Item).annotations_link))

        _json.update({'annotations': self.annotations_link,
                      'annotationsCount': self.annotations_count,
                      'dataset': self.dataset_url})
        return _json

    def download(
            self,
            # download options
            local_path=None,
            file_types=None,
            save_locally=True,
            num_workers=None,
            annotation_options=None,
            overwrite=False,
            to_items_folder=True,
            thickness=1,
            with_text=False
    ):
        """
        Download dataset by filters.
        Filtering the dataset for items and save them local
        Optional - also download annotation, mask, instance and image mask of the item

        :param local_path: local folder or filename to save to.
        :param to_items_folder: Create 'items' folder and download items to it
        :param overwrite: optional - default = False
        :param file_types: a list of file type to download. e.g ['video/webm', 'video/mp4', 'image/jpeg', 'image/png']
        :param num_workers: default - 32
        :param save_locally: bool. save to disk or return a buffer
        :param annotation_options: download annotations options: dl.ViewAnnotationOptions.list()
        :param with_text: optional - add text to annotations, default = False
        :param thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :return: Output (list)
        """
        # if dir - concatenate local path and item name
        if local_path is not None:
            _, ext = os.path.splitext(local_path)
            if not ext:
                local_path = os.path.join(local_path, self.name)

        # download
        return self.items.download(items=self,
                                   local_path=local_path,
                                   file_types=file_types,
                                   save_locally=save_locally,
                                   num_workers=num_workers,
                                   annotation_options=annotation_options,
                                   overwrite=overwrite,
                                   to_items_folder=to_items_folder,
                                   thickness=thickness,
                                   with_text=with_text)

    def delete(self):
        """
        Delete item from platform
        :return: True
        """
        return self.items.delete(item_id=self.id)

    def update(self, system_metadata=False):
        """
        Update items metadata
        :param system_metadata: bool
        :return: Item object
        """
        return self.items.update(item=self, system_metadata=system_metadata)

    def move(self, new_path):
        """
        Move item from one folder to another in Platform
        :param new_path: new path to move item to. Format: /main_folder/sub_folder/.../item_name.type
        :return: True
        """
        assert isinstance(new_path, str)
        if not new_path.startswith('/'):
            new_path = '/' + new_path
        if '.' in new_path:
            if len(new_path.split('.')) > 2:
                raise exceptions.PlatformException('400', 'Remote path cannot include dots')
            else:
                self.filename = new_path
        else:
            self.filename = new_path + '/' + self.name

        return self.update(system_metadata=True)

    def clone(self, dst_dataset_id, remote_filepath=None, metadata=None,
              with_annotations=True, with_metadata=True, with_task_annotations_status=False):
        if remote_filepath is None:
            remote_filepath = self.filename
        return self.items.clone(item_id=self.id,
                                dst_dataset_id=dst_dataset_id,
                                remote_filepath=remote_filepath,
                                metadata=metadata,
                                with_annotations=with_annotations,
                                with_metadata=with_metadata,
                                with_task_annotations_status=with_task_annotations_status)

    def open_in_web(self):
        self.items.open_in_web(item=self)

    def change_status(self, status):
        # TODO - deprecate
        logger.warning('[DeprecationWarning] "change_status(status)" method will be deprecated after version 1.17.0\n'
                       'Please use method "update_status(status)"')
        self.update_status(status=status)

    def update_status(self, status):
        if status not in ['completed', 'approved', 'discarded']:
            raise exceptions.PlatformException('400',
                                               'Unknown status: {}. Please chose from: completed, approved, discarded'.format(
                                                   status))
        try:
            annotation_definition = entities.Classification(label=status)
            entities.Annotation.new(item=self, annotation_definition=annotation_definition,
                                    metadata={'system': {'system': True}}).upload()
            return True
        except Exception:
            logger.error('Error updating status. Please use platform')
            logger.debug(traceback.format_exc())
            return False


class ModalityTypeEnum:
    """
    State enum
    """

    OVERLAY = "overlay"


class Modality:
    def __init__(self, _json=None, modality_type=None, ref=None, ref_type='id', name=None):
        if _json is None:
            _json = dict()
        self.type = _json.get('type', modality_type)
        self.ref_type = _json.get('refType', ref_type)
        self.ref = _json.get('ref', ref)
        self.name = _json.get('name', name)

    def to_json(self):
        return {"type": self.type,
                "ref": self.ref,
                "refType": self.ref_type,
                "name": self.name}


class Modalities:
    def __init__(self, item):
        assert isinstance(item, Item)
        self.item = item

    @property
    def modalities(self):
        return self.item.metadata.get('modalities', None)

    def create(self, name, ref, ref_type='id', modality_type=ModalityTypeEnum.OVERLAY):
        if self.modalities is None:
            self.item.metadata['modalities'] = list()

        _json = {
            "type": modality_type,
            "refType": ref_type,
            "ref": ref,
            "name": name
        }

        self.item.metadata['modalities'].append(_json)

        return Modality(_json=_json)

    def delete(self, name):
        if self.modalities is not None:
            for modality in self.item.metadata['modalities']:
                if name == modality['name']:
                    self.item.metadata['modalities'].remove(modality)
                    return Modality(_json=modality)
        return None

    def list(self):
        modalities = list()
        if self.modalities is not None:
            modalities = list()
            for modality in self.item.metadata['modalities']:
                modalities.append(Modality(_json=modality))
        return modalities



import threading
import asyncio
import logging

import io

logger = logging.getLogger(name=__name__)


class AsyncThreadEventLoop(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, loop, n, name, *args, **kwargs):
        super(AsyncThreadEventLoop, self).__init__(*args, **kwargs)
        self.loop = loop
        self.n = n
        self.name = name
        self._count = 0
        self._semaphores = dict()

    def count_up(self):
        self._count += 1
        # logger.debug('eventloop {} with {} tasks'.format(self.name, self._count))

    def count_down(self):
        self._count -= 1
        # logger.debug('eventloop {} with {} tasks'.format(self.name, self._count))

    def run(self):
        logger.debug('Starting event loop "{}" with bounded semaphore to {}'.format(self.name, self.n))
        
        def exception_handler(loop, context):
            logger.debug("[Asyc] EventLoop: {} caught the following exception: {}".format(self.name, context['message']))

        self.loop.set_exception_handler(exception_handler)
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()
        logger.debug('Ended event loop "{}" with bounded semaphore to {}'.format(self.name, self.n))

    def semaphore(self, name):
        if name not in self._semaphores:
            self._semaphores[name] = asyncio.Semaphore(self.n, loop=self.loop)
        return self._semaphores[name]

    def stop(self):
        self.loop.call_soon_threadsafe(self.loop.stop)  # here


class AsyncResponse:
    def __init__(self, text, _json, async_resp):
        self.text = text
        self._json = _json
        self.async_resp = async_resp

    def json(self):
        return self._json

    @property
    def status_code(self):
        return self.async_resp.status

    @property
    def reason(self):
        return self.async_resp.reason

    @property
    def ok(self):
        return self.async_resp.status < 400

    @property
    def request(self):
        return self.async_resp.request_info

    @property
    def headers(self):
        return self.async_resp.headers


class DummyErrorResponse:
    def __init__(self, error, trace):
        self.status = 3001  # SDK Error
        self.reason = trace
        self.error = error
        self.request_info = {}
        self.headers = {}


class AsyncResponseError(AsyncResponse):
    def __init__(self, error, trace):
        async_resp = DummyErrorResponse(error=error, trace=trace)
        _json = {'error': error}
        text = error
        super().__init__(async_resp=async_resp,
                         _json=_json,
                         text=text)


class AsyncUploadStream(io.IOBase):
    def __init__(self, buffer, callback=None):
        self.buffer = buffer
        self.buffer.seek(0)
        self.callback = callback

    def read(self, size):
        if self.callback is not None:
            self.callback(size)
        return self.buffer.read(size)



from .. import miscellaneous


class BaseEntity(object):
    is_fetched = True

    def print(self, to_return=False):
        return miscellaneous.List([self]).print(to_return=to_return)

    def to_df(self, show_all=False):
        return miscellaneous.List([self]).to_df(show_all=show_all)

    # def __getattribute__(self, attr):
    #     if super(BaseEntity, self).__getattribute__(attr) is None:
    #         pass
    #     return super(BaseEntity, self).__getattribute__(attr)



import threading
import logging

from .. import exceptions, entities, repositories, miscellaneous, services

logger = logging.getLogger(name=__name__)


class Executions:
    """
    Service Executions repository
    """

    def __init__(self,
                 client_api: services.ApiClient,
                 service: entities.Service = None,
                 project: entities.Project = None):
        self._client_api = client_api
        self._service = service
        self._project = project

    ############
    # entities #
    ############
    @property
    def service(self) -> entities.Service:
        if self._service is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "service". need to set a Service entity or use service.executions repository')
        assert isinstance(self._service, entities.Service)
        return self._service

    @service.setter
    def service(self, service: entities.Service):
        if not isinstance(service, entities.Service):
            raise ValueError('Must input a valid Service entity')
        self._service = service

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            if self._service is not None:
                self._project = self._service._project
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use Project.executions repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    def __get_from_entity(self, name, value):
        project_id = None
        try:
            from_dataset = False
            entity_obj = None
            for input_type in entities.FunctionIO.INPUT_TYPES:
                entity = input_type.lower()
                param = '{}_id'.format(entity)
                if isinstance(value, dict):
                    if param in value:
                        repo = getattr(repositories, '{}s'.format(input_type))(client_api=self._client_api)
                        entity_obj = repo.get(**{param: value[param]})
                        if param in ['annotation_id', 'item_id']:
                            from_dataset = True
                            if param == 'item_id':
                                entity_obj = entity_obj.dataset
                            else:
                                entity_obj = repositories.Datasets(client_api=self._client_api).get(
                                    dataset_id=entity_obj.dataset_id)
                        elif param in ['dataset_id']:
                            from_dataset = True
                        break
                elif isinstance(value, str):
                    if entity == name:
                        repo = getattr(repositories, '{}s'.format(input_type))(client_api=self._client_api)
                        entity_obj = repo.get(**{param: value})
                        if name in ['annotation', 'item']:
                            from_dataset = True
                            if param == 'item_id':
                                entity_obj = entity_obj.dataset
                            else:
                                entity_obj = repositories.Datasets(client_api=self._client_api).get(
                                    dataset_id=entity_obj.dataset_id)
                        elif name in ['dataset']:
                            from_dataset = True
                        break

            if entity_obj is not None:
                if isinstance(entity_obj, entities.Project):
                    project_id = entity_obj.id
                elif from_dataset:
                    project_id = entity_obj.projects[0]
                else:
                    project_id = entity_obj.project_id
        except Exception:
            project_id = None

        return project_id

    def __get_project_id(self, project_id=None, payload=None):
        if project_id is None:
            inputs = payload.get('input', dict())
            try:
                for key, val in inputs.items():
                    project_id = self.__get_from_entity(name=key, value=val)
                    if project_id is not None:
                        break
            except Exception:
                project_id = None

            if project_id is None:
                # if still None - get from current repository
                if self._project is not None:
                    project_id = self._project.id
                else:
                    raise exceptions.PlatformException('400', 'Please provide project_id')

        return project_id

    ###########
    # methods #
    ###########
    def create(self,
               # executions info
               service_id=None, execution_input=None, function_name=None,
               # inputs info
               resource=None, item_id=None, dataset_id=None, annotation_id=None, project_id=None,
               # execution config
               sync=False, stream_logs=False, return_output=False) -> entities.Execution:
        """
        Execute a function on an existing service

        :param service_id: service id to execute on
        :param function_name: function name to run
        :param project_id: resource's project
        :param execution_input: input dictionary or list of FunctionIO entities
        :param dataset_id: optional - input to function
        :param item_id: optional - input to function
        :param annotation_id: optional - input to function
        :param resource: input type.
        :param sync: wait for function to end
        :param stream_logs: prints logs of the new execution. only works with sync=True
        :param return_output: if True and sync is True - will return the output directly
        :return:
        """
        if service_id is None:
            if self._service is None:
                raise exceptions.PlatformException('400', 'Please provide service id')
            service_id = self._service.id

        if resource is None:
            if annotation_id is not None:
                resource = entities.PackageInputType.ANNOTATION
            elif item_id is not None:
                resource = entities.PackageInputType.ITEM
            elif dataset_id is not None:
                resource = entities.PackageInputType.DATASET

        # payload
        payload = dict()
        if execution_input is None:
            if resource is not None:
                inputs = {resource.lower(): {
                    'dataset_id': dataset_id}
                }
                if item_id is not None:
                    inputs[resource.lower()]['item_id'] = item_id
                if annotation_id is not None:
                    inputs[resource.lower()]['annotation_id'] = annotation_id
                payload['input'] = inputs
        else:
            if isinstance(execution_input, dict):
                payload['input'] = execution_input
            else:
                if not isinstance(execution_input, list):
                    execution_input = [execution_input]
                if len(execution_input) > 0 and isinstance(execution_input[0], entities.FunctionIO):
                    payload['input'] = dict()
                    for single_input in execution_input:
                        payload['input'].update(single_input.to_json(resource='execution'))
                else:
                    raise exceptions.PlatformException('400', 'Unknown input type')

        payload['projectId'] = self.__get_project_id(project_id=project_id, payload=payload)

        if function_name is not None:
            payload['functionName'] = function_name
        else:
            payload['functionName'] = entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME

        # request url
        url_path = '/executions/{service_id}'.format(service_id=service_id)
        if sync and not return_output and not stream_logs:
            url_path += '?sync=true'
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url_path,
                                                         json_req=payload)
        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        execution = entities.Execution.from_json(_json=response.json(),
                                                 client_api=self._client_api,
                                                 service=self._service)

        if sync and (stream_logs or return_output):
            thread = None
            if stream_logs:
                thread = threading.Thread(target=self.logs,
                                          kwargs={'execution_id': execution.id,
                                                  'follow': True,
                                                  'until_completed': True})
                thread.start()
            execution = self.get(execution_id=execution.id,
                                 sync=True)
            # stream logs
            if stream_logs and thread is not None:
                thread.join()
        if sync and return_output:
            return execution.output
        return execution

    def _list(self, filters: entities.Filters):
        """
        List service executions
        :return:
        """
        url = '/query/FaaS'

        # request
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List service executions
        :return:
        """
        # default filtersf
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.EXECUTION)
            if self._project is not None:
                filters.add(field='projectId', values=self._project.id)
            if self._service is not None:
                filters.add(field='serviceId', values=self._service.id)

        # assert type filters
        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400', message='Unknown filters type')

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Execution]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return execution list
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.apply_async(entities.Execution._protected_from_json,
                                            kwds={'client_api': self._client_api,
                                                  '_json': item,
                                                  'service': self._service})
        # wait for all jobs
        _ = [j.wait() for j in jobs]
        # get results
        results = [j.get() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    def get(self, execution_id=None, sync=False) -> entities.Execution:
        """
        Get Service execution object

        :param execution_id:
        :param sync: wait for the execution to finish
        :return: Service execution object
        """
        url_path = "/executions/{}".format(execution_id)
        if sync:
            url_path += '?sync=true'

        success, response = self._client_api.gen_request(req_type="get",
                                                         path=url_path)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Execution.from_json(client_api=self._client_api,
                                            _json=response.json(),
                                            service=self._service)

    def logs(self, execution_id, follow=True, until_completed=True):
        """

        """
        logs = self.service.log(execution_id=execution_id, follow=follow)
        end_string = '[Done] Executing function.'
        try:
            for log in logs:
                print(log)
                if until_completed and end_string in log:
                    break
        except KeyboardInterrupt:
            pass

    def increment(self, execution: entities.Execution):
        """
        Increment attempts

        :return: int
        """
        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/executions/{}/attempts'.format(execution.id))

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        else:
            return response.json()

    def wait(self, execution_id):
        """
        Get Service execution object

        :param execution_id:
        :return: Service execution object
        """
        url_path = "/executions/{}?sync=true".format(execution_id)
        while True:
            success, response = self._client_api.gen_request(req_type="get",
                                                             path=url_path,
                                                             log_error=False)
            if response.status_code in [500, 504]:
                # while timing out continue wait
                pass
            else:
                break
        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        # return entity
        return entities.Execution.from_json(client_api=self._client_api,
                                            _json=response.json(),
                                            service=self._service)

    def terminate(self, execution: entities.Execution):
        """
        Terminate Execution

        :return:
        """
        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/executions/{}/terminate'.format(execution.id))

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        else:
            return entities.Execution.from_json(_json=response.json(),
                                                service=self._service,
                                                client_api=self._client_api)

    def update(self, execution: entities.Execution) -> entities.Execution:
        """
        Update execution changes to platform
        :param execution: execution entity
        :return: execution entity
        """
        # payload
        payload = execution.to_json()

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='/executions/{}'.format(execution.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        if self._service is not None:
            service = self._service
        else:
            service = execution._service

        return entities.Execution.from_json(_json=response.json(),
                                            service=service,
                                            client_api=self._client_api)

    def progress_update(self, execution_id, status=None, percent_complete=None, message=None, output=None):
        """
        Update Execution Progress

        :param execution_id:
        :param status:
        :param percent_complete:
        :param message:
        :param output:
        :return:
        """
        # create payload
        payload = dict()
        if status is not None:
            payload['status'] = status
        else:
            if percent_complete is not None and isinstance(percent_complete, int):
                if percent_complete < 100:
                    payload['status'] = 'inProgress'
                else:
                    payload['status'] = 'completed'
            elif output is not None:
                payload['status'] = 'completed'
            else:
                payload['status'] = 'inProgress'

        if percent_complete is not None:
            payload['percentComplete'] = percent_complete
        if message is not None:
            payload['message'] = message
        if output is not None:
            payload['output'] = output

        # request
        success, response = self._client_api.gen_request(
            req_type="post",
            path="/executions/{}/progress".format(execution_id),
            json_req=payload
        )

        # exception handling
        if success:
            return entities.Execution.from_json(_json=response.json(),
                                                client_api=self._client_api,
                                                service=self._service)
        else:
            raise exceptions.PlatformException(response)



import logging

import jwt

from .. import entities, miscellaneous, exceptions, services

logger = logging.getLogger(name=__name__)


class Projects:
    """
    Projects repository
    """

    def __init__(self, client_api: services.ApiClient):
        self._client_api = client_api

    def __get_from_cache(self) -> entities.Project:
        project = self._client_api.state_io.get('project')
        if project is not None:
            project = entities.Project.from_json(_json=project, client_api=self._client_api)
        return project

    def __get_by_id(self, project_id: str) -> entities.Project:
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/projects/{}'.format(project_id))
        if success:
            project = entities.Project.from_json(client_api=self._client_api,
                                                 _json=response.json())
        else:
            # raise PlatformException(response)
            # TODO because of a bug in gate wrong error is returned so for now manually raise not found
            raise exceptions.PlatformException(error="404", message="Project not found")
        return project

    def __get_by_identifier(self, identifier=None) -> entities.Project:
        projects = self.list()
        projects_by_name = [project for project in projects if identifier in project.id or identifier in project.name]
        if len(projects_by_name) == 1:
            return projects_by_name[0]
        elif len(projects_by_name) > 1:
            raise Exception('Multiple projects with this name/identifier exist')
        else:
            raise Exception("Project not found")

    def open_in_web(self, project_name: str = None, project_id: str = None, project: entities.Project = None):
        if project_id is None:
            if project is not None:
                project_id = project.id
            elif project_name is not None:
                project_id = self.get(project_name=project_name)
            else:
                raise exceptions.PlatformException('400', 'Please provide project, project name or project id')
        self._client_api._open_in_web(resource_type='project', project_id=project_id)

    def checkout(self,
                 identifier: str = None,
                 project_name: str = None,
                 project_id: str = None,
                 project: entities.Project = None):
        """
        Check-out a project
        :param project:
        :param project_id:
        :param project_name:
        :param identifier: project name or partial id
        :return:
        """
        if project is None:
            if project_id is not None or project_name is not None:
                project = self.get(project_id=project_id, project_name=project_name)
            elif identifier is not None:
                project = self.__get_by_identifier(identifier=identifier)
            else:
                raise exceptions.PlatformException(error='400',
                                                   message='Must provide partial/full id/name to checkout')
        self._client_api.state_io.put('project', project.to_json())
        logger.info('Checked out to project {}'.format(project.name))

    def _send_mail(self, project_id: str, send_to: str, title: str, content: str) -> bool:
        url = '/projects/{}/mail'.format(project_id)
        assert isinstance(title, str)
        assert isinstance(content, str)
        if self._client_api.token is not None:
            sender = jwt.decode(self._client_api.token, algorithms=['HS256'], verify=False)['email']
        else:
            raise exceptions.PlatformException('600', 'Token expired please log in')

        payload = {
            'to': send_to,
            'from': sender,
            'subject': title,
            'body': content
        }

        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=payload)

        if not success:
            raise exceptions.PlatformException(response)
        return True

    def add_member(self, email: str, project_id: str, role: str = 'engineer'):
        url_path = '/projects/{}/members/{}'.format(project_id, email)
        payload = dict(role=role)
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url_path,
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    def list(self) -> miscellaneous.List[entities.Project]:
        """
        Get users project's list.
        :return: List of Project objects
        """
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/projects')

        if success:
            pool = self._client_api.thread_pools(pool_name='entity.create')
            projects_json = response.json()
            jobs = [None for _ in range(len(projects_json))]
            # return triggers list
            for i_project, project in enumerate(projects_json):
                jobs[i_project] = pool.apply_async(entities.Project._protected_from_json,
                                                   kwds={'client_api': self._client_api,
                                                         '_json': project})
            # wait for all jobs
            _ = [j.wait() for j in jobs]
            # get all results
            results = [j.get() for j in jobs]
            # log errors
            _ = [logger.warning(r[1]) for r in results if r[0] is False]
            # return good jobs
            projects = miscellaneous.List([r[1] for r in results if r[0] is True])
        else:
            logger.error('Platform error getting projects')
            raise exceptions.PlatformException(response)
        return projects

    def get(self,
            project_name: str = None,
            project_id: str = None,
            checkout: bool = False,
            fetch: bool = None) -> entities.Project:
        """
        Get a Project object
        :param checkout:
        :param project_name: optional - search by name
        :param project_id: optional - search by id
        :param fetch: optional - fetch entity from platform, default taken from cookie
        :return: Project object

        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if project_id is None and project_name is None:
            project = self.__get_from_cache()
            if project is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='Checked out not found, must provide either project id or project name')
        elif fetch:
            if project_id is not None:
                project = self.__get_by_id(project_id)
            elif project_name is not None:
                projects = self.list()
                project = [project for project in projects if project.name == project_name]
                if not project:
                    # list is empty
                    raise exceptions.PlatformException(error='404',
                                                       message='Project not found. Name: {}'.format(project_name))
                    # project = None
                elif len(project) > 1:
                    # more than one matching project
                    raise exceptions.PlatformException(
                        error='404',
                        message='More than one project with same name. Please "get" by id')
                else:
                    project = project[0]
            else:
                raise exceptions.PlatformException(
                    error='404',
                    message='No input and no checked-out found')
        else:
            project = entities.Project.from_json(_json={'id': project_id,
                                                        'name': project_name},
                                                 client_api=self._client_api,
                                                 is_fetched=False)
        assert isinstance(project, entities.Project)
        if checkout:
            self.checkout(project=project)
        return project

    def delete(self,
               project_name: str = None,
               project_id: str = None,
               sure: bool = False,
               really: bool = False) -> bool:
        """
        Delete a project forever!
        :param project_name: optional - search by name
        :param project_id: optional - search by id
        :param sure: are you sure you want to delete?
        :param really: really really?

        :return: True
        """
        if sure and really:
            if project_id is None:
                project = self.get(project_name=project_name)
                project_id = project.id
            success, response = self._client_api.gen_request(req_type='delete',
                                                             path='/projects/{}'.format(project_id))
            if not success:
                raise exceptions.PlatformException(response)
            logger.info('Project id {} deleted successfully'.format(project_id))
            return True
        else:
            raise exceptions.PlatformException(
                error='403',
                message='Cant delete project from SDK. Please login to platform to delete')

    def update(self,
               project: entities.Project,
               system_metadata: bool = False) -> entities.Project:
        """
        Update a project
        :return: Project object
        """
        url_path = '/projects/{}'.format(project.id)
        if system_metadata:
            url_path += '?system=true'
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=project.to_json())
        if success:
            return project
        else:
            raise exceptions.PlatformException(response)

    def create(self,
               project_name: str,
               checkout: bool = False) -> entities.Project:
        """
        Create a new project
        :param checkout:
        :param project_name:
        :return: Project object
        """
        payload = {'name': project_name}
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/projects',
                                                         data=payload)
        if success:
            project = entities.Project.from_json(client_api=self._client_api,
                                                 _json=response.json())
        else:
            raise exceptions.PlatformException(response)
        assert isinstance(project, entities.Project)
        if checkout:
            self.checkout(project=project)
        return project



import threading
import requests
import logging
import hashlib
import base64
import json
import jwt
import os
from urllib.parse import urlencode

from .service_defaults import DEFAULT_ENVIRONMENTS

logger = logging.getLogger(name=__name__)
threadLock = threading.Lock()


def login_m2m(api_client, email, password, client_id, client_secret, force=False):
    login_secret(api_client=api_client,
                 email=email,
                 password=password,
                 client_id=client_id,
                 client_secret=client_secret,
                 force=force)


def login_secret(api_client, email, password, client_id, client_secret, force=False):
    """
    Login with email and password from environment variables
    :param api_client: ApiClient instance
    :param email: user email. if already logged in with same user - login will NOT happen. see "force"
    :param password: user password
    :param client_id:
    :param client_secret:
    :param force: force login. in case login with same user but want to get a new JWT
    :return:
    """
    # check if already logged in with SAME email
    if api_client.token is not None or api_client.token == '':
        try:
            payload = jwt.decode(api_client.token, algorithms=['HS256'], verify=False)
            if 'email' in payload and \
                    payload['email'] == email and \
                    not api_client.token_expired() and \
                    not force:
                logger.warning('Trying to login with same email but token not expired. Not doing anything... '
                               'Set "force" flag to True to login anyway.')
                return True
        except jwt.exceptions.DecodeError:
            logger.debug('{}'.format('Cant decode token. Force login is used'))

    logger.info('[Start] Login Secret')
    environment = api_client.environment
    audience = None
    auth0_url = None
    for env, env_params in api_client.environments.items():
        if env == environment:
            audience = env_params['audience']
            auth0_url = env_params['auth0_url']
    missing = False
    if audience is None:
        logger.error('audience not found. Please add a new environment to SDK. env: {}'.format(environment))
        missing = True
    if auth0_url is None:
        logger.error('auth0_url not found. Please add a new environment to SDK. env: {}'.format(environment))
        missing = True
    if missing:
        raise ConnectionError('Some values are missing. See above for full error')
    # need to login
    payload = {'username': email,
               'password': password,
               'grant_type': 'password',
               'audience': audience,
               'scope': 'openid email offline_access',
               'client_id': client_id,
               'client_secret': client_secret
               }
    headers = {'content-type': 'application/json'}
    token_url = auth0_url + '/oauth/token'
    resp = requests.request("POST", token_url, data=json.dumps(payload), headers=headers)
    if not resp.ok:
        api_client.print_bad_response(resp)
        return False
    else:
        response_dict = resp.json()
        api_client.token = response_dict['id_token']  # this will also set the refresh_token to None
        if 'refresh_token' in response_dict:
            api_client.refresh_token = response_dict['refresh_token']
        # set new client id for refresh
        api_client.environments[environment]['client_id'] = client_id
        payload = jwt.decode(api_client.token, algorithms=['HS256'], verify=False)
        if 'email' in payload:
            logger.info('[Done] Login Secret. User: {}'.format(payload['email']))
        else:
            logger.info('[Done] Login Secret. User: {}'.format(email))
            logger.info(payload)
    return True


def login(api_client, audience=None, auth0_url=None, client_id=None):
    """
    Login using Auth0.
    :param api_client: ApiClient instance
    :param audience:
    :param auth0_url:
    :param client_id:
    :return:
    """
    import webbrowser
    from http.server import BaseHTTPRequestHandler, HTTPServer
    from urllib.parse import urlparse, parse_qs

    logger.info('Logging in to Dataloop...')
    login_success = False
    # create a Code Verifier
    n_bytes = 64
    verifier = base64.urlsafe_b64encode(os.urandom(n_bytes)).rstrip(b'=')
    # https://tools.ietf.org/html/rfc7636#section-4.1
    # minimum length of 43 characters and a maximum length of 128 characters.
    if len(verifier) < 43:
        raise ValueError("Verifier too short. n_bytes must be > 30.")
    elif len(verifier) > 128:
        raise ValueError("Verifier too long. n_bytes must be < 97.")
    # Create a code challenge
    digest = hashlib.sha256(verifier).digest()
    challenge = base64.urlsafe_b64encode(digest).rstrip(b'=')

    ################################
    # auth0 parameters for request #
    ################################
    # get env from url
    if api_client.environment in api_client.environments.keys():
        env_params = DEFAULT_ENVIRONMENTS[api_client.environment]
        audience = env_params['audience']
        client_id = env_params['client_id']
        auth0_url = env_params['auth0_url']
    else:
        missing = False
        if audience is None:
            logger.error('Missing parameter for environment missing. Need to input "audience"')
            missing = True
        if client_id is None:
            logger.error('Missing parameter for environment missing. Need to input "client_id"')
            missing = True
        if auth0_url is None:
            logger.error('Missing parameter for environment missing. Need to input "auth0_url"')
            missing = True
        if missing:
            raise ConnectionError('Missing parameter for environment. see above')
        # add to login parameters
        api_client.add_environment(environment=api_client.environment,
                                   audience=audience,
                                   client_id=client_id,
                                   auth0_url=auth0_url)

    redirect_url = 'http://localhost:3001/token'

    # set url request for auth0
    payload = {'code_challenge_method': 'S256',
               'code_challenge': challenge,
               'response_type': 'code',
               'audience': audience,
               'scope': 'openid email offline_access',
               'client_id': client_id,
               'redirect_uri': redirect_url}

    query_string = urlencode(payload, doseq=True)

    # set up local server to get response from auth0
    global query_dict
    query_dict = None

    class RequestHandler(BaseHTTPRequestHandler):

        def log_message(api_client, format, *args):
            return

        def do_GET(api_client):
            global query_dict
            parsed_path = urlparse(api_client.path)
            query_dict = parse_qs(parsed_path.query)
            try:
                # working directory when running from command line
                location = os.path.dirname(os.path.realpath(__file__))
            except NameError:
                # working directory when running from console
                location = './dtlpy/services'
            filename = os.path.join(location, '..', 'assets', 'lock_open.png')
            if query_dict and 'code' in query_dict:
                if os.path.isfile(filename):
                    with open(filename, 'rb') as f:
                        # Open the static file requested and send it
                        api_client.send_response(200)
                        api_client.send_header('Content-type', 'image/jpg')
                        api_client.end_headers()
                        api_client.wfile.write(f.read())
            return

    port = 3001
    # print('Listening on localhost:%s' % port)
    server = HTTPServer(('', port), RequestHandler)
    # set timeout to 1min (waiting for user to login)
    server.timeout = 60  # timeout 1 min
    try:
        # open browser to Auth0 login page
        webbrowser.open(url=auth0_url + '/authorize' + '?%s' % query_string, new=2, autoraise=True)
        # wait for request
        server.handle_request()
        # check the global list for the token
        if query_dict and 'code' in query_dict:
            # authentication code received from auth0 - can continue login
            # payload for auth0 token request
            payload = {'grant_type': 'authorization_code',
                       'client_id': client_id,
                       'code_verifier': verifier.decode(),
                       'code': query_dict['code'][0],
                       'redirect_uri': redirect_url}
            resp = requests.request("POST",
                                    auth0_url + '/oauth/token',
                                    json=payload,
                                    headers={'content-type': 'application/json'})
            if not resp.ok:
                api_client.print_bad_response(resp)
                login_success = False
            else:
                response_dict = resp.json()
                final_token = response_dict['id_token']
                api_client.token = final_token  # this will also set the refresh_token to None
                if 'refresh_token' in response_dict:
                    api_client.refresh_token = response_dict['refresh_token']
                api_client.environments[api_client.environment]['client_id'] = client_id
                payload = jwt.decode(api_client.token, algorithms=['HS256'], verify=False)
                if 'email' in payload:
                    logger.info('Logged in: %s' % payload['email'])
                else:
                    logger.info('Logged in: unknown user')
                login_success = True
        else:
            # if time out passed (in seconds) break
            logger.error('Timeout reached: getting token from server')
            raise ConnectionError('Timeout reached: getting token from server')
    except Exception as err:
        logger.exception('Error in http server for getting token')
    finally:
        # shutdown local server
        server.server_close()
    return login_success



from . import BaseAnnotationDefinition


class Subtitle(BaseAnnotationDefinition):
    """
        Subtitle annotation object
    """

    def __init__(self, text, label, attributes=None):
        self.type = "subtitle"
        self.text = text
        self.label = label
        if attributes is None:
            attributes = list()
        self.attributes = attributes

    def to_coordinates(self, color):
        return {"text": self.text}

    @staticmethod
    def from_coordinates(coordinates):
        return coordinates["text"]

    @classmethod
    def from_json(cls, _json):
        attributes = _json.get("attributes", list())
        if "coordinates" in _json:
            text = cls.from_coordinates(coordinates=_json["coordinates"])
        elif "data" in _json:
            text = cls.from_coordinates(coordinates=_json["data"])
        else:
            raise ValueError('Bad json, "coordinates" or "data" not found')
        return cls(
            text=text,
            label=_json["label"],
            attributes=attributes
        )



import os
import datetime
import json
import numpy as np


class Reporter:
    """
     ThreadPool Report summary
    """
    ITEMS_DOWNLOAD = 'downloader'
    ITEMS_UPLOAD = 'uploader'
    CONVERTER = 'converter'

    def __init__(self, num_workers, resource):
        self._num_workers = num_workers
        self._refs = [None for _ in range(num_workers)]
        self._success = [False for _ in range(num_workers)]
        self._status = ["" for _ in range(num_workers)]
        self._errors = [None for _ in range(num_workers)]
        self._output = [None for _ in range(num_workers)]
        self._resource = resource

    @property
    def has_errors(self):
        return self.failure_count > 0

    @property
    def output(self):
        output = [output for output in self._output if output is not None]
        return output[0] if len(output) == 1 else output

    @property
    def status_list(self):
        return np.unique(self._status)

    @property
    def num_workers(self):
        return self._num_workers

    @property
    def failure_count(self):
        return len([suc for suc in self._success if suc is False])

    @property
    def success_count(self):
        return len([suc for suc in self._success if suc is True])

    def status_count(self, status):
        return self._status.count(status)

    def set_index(self, i_item, error=None, status=None, ref=None, success=None, output=None):
        if error is not None:
            self._errors[i_item] = error
        if status is not None:
            self._status[i_item] = status
        if ref is not None:
            self._refs[i_item] = ref
        if success is not None:
            self._success[i_item] = success
        if success and output is not None:
            self._output[i_item] = output

    def generate_log_files(self):
        log_filepath = os.path.join(os.getcwd(),
                                    "log_{}_{}.json".format(self._resource,
                                                            datetime.datetime.utcnow().strftime("%Y%m%d_%H%M%S")))
        errors_list = [self._errors[i_job] for i_job, suc in enumerate(self._success) if suc is False]
        ref_list = [self._refs[i_job] for i_job, suc in enumerate(self._success) if suc is False]
        errors_json = {item_ref: error for item_ref, error in zip(ref_list, errors_list)}
        with open(log_filepath, "w") as f:
            json.dump(errors_json, f, indent=2)

        return log_filepath



import attr
import logging
from .. import repositories, exceptions, entities

logger = logging.getLogger(name=__name__)


@attr.s
class Assignment(entities.BaseEntity):
    """
    Assignment object
    """

    # platform
    name = attr.ib()
    annotator = attr.ib()
    status = attr.ib()
    project_id = attr.ib()
    metadata = attr.ib(repr=False)
    id = attr.ib()
    url = attr.ib(repr=False)

    # sdk
    _client_api = attr.ib(repr=False)
    _task = attr.ib(default=None, repr=False)
    _assignments = attr.ib(default=None, repr=False)
    _project = attr.ib(default=None, repr=False)
    _dataset = attr.ib(default=None, repr=False)
    _datasets = attr.ib(default=None, repr=False)

    @classmethod
    def from_json(cls, _json, client_api, project=None, task=None, dataset=None):
        return cls(
            name=_json.get('name', None),
            annotator=_json.get('annotator', None),
            status=_json.get('status', None),
            project_id=_json.get('projectId', None),
            metadata=_json.get('metadata', dict()),
            url=_json.get('url', None),
            id=_json['id'],
            client_api=client_api,
            project=project,
            dataset=dataset,
            task=task
        )

    @property
    def assignments(self):
        if self._assignments is None:
            self._assignments = repositories.Assignments(client_api=self._client_api, project=self.project,
                                                         task=self._task, dataset=self._dataset)
        assert isinstance(self._assignments, repositories.Assignments)
        return self._assignments

    @property
    def project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id,
                                                                                   fetch=None)

        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def dataset_id(self):
        dataset_id = self.metadata.get('datasetId', None)
        if dataset_id is None:
            system_metadata = self.metadata.get('system', dict())
            dataset_id = system_metadata.get('datasetId', None)
        return dataset_id

    @property
    def datasets(self):
        if self._datasets is None:
            self._datasets = repositories.Datasets(client_api=self._client_api, project=self._project)
        assert isinstance(self._datasets, repositories.Datasets)
        return self._datasets

    @property
    def dataset(self):
        if self._dataset is None:
            self._dataset = self.datasets.get(dataset_id=self.dataset_id)
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Assignment)._client_api,
                                                              attr.fields(Assignment)._project,
                                                              attr.fields(Assignment).project_id,
                                                              attr.fields(Assignment)._assignments,
                                                              attr.fields(Assignment)._dataset,
                                                              attr.fields(Assignment)._task))
        _json['projectId'] = self.project_id
        return _json

    def update(self, system_metadata=False):
        return self.assignments.update(assignment=self, system_metadata=system_metadata)

    def assign_items(self, dataset=None, filters=None, items=None):
        """

        :param filters:
        :param dataset:
        :param items:
        :return:
        """
        if dataset is None:
            dataset = self.dataset

        return self.assignments.assign_items(dataset=dataset, assignment_id=self.id, filters=filters, items=items)

    def remove_items(self, dataset=None, filters=None, items=None):
        """

        :param dataset:
        :param filters:
        :param items:
        :return:
        """
        if dataset is None:
            dataset = self.dataset

        return self.assignments.remove_items(dataset=dataset, assignment_id=self.id, filters=filters, items=items)

    def get_items(self, dataset=None, filters=None):
        """

        :param dataset:
        :return:
        """
        if dataset is None:
            dataset = self.dataset

        return self.assignments.get_items(dataset=dataset, assignment=self, filters=filters)

    def reassign(self, assignee_id):
        """
        Reassign an assignment
        :return: Assignment object
        """
        return self.assignments.reassign(assignment=self, task=self._task, assignee_id=assignee_id)

    def redistribute(self, workload):
        """
        Redistribute an assignment
        :return: Assignment object
        """
        return self.assignments.redistribute(assignment=self, task=self._task, workload=workload)


@attr.s
class WorkloadUnit:
    """
    WorkloadUnit object
    """

    # platform
    assignee_id = attr.ib(type=str)
    load = attr.ib(type=float, default=0)

    # noinspection PyUnusedLocal
    @load.validator
    def check_load(self, attribute, value):
        if value < 0 or value > 100:
            raise exceptions.PlatformException('400', 'Value must be a number between 0 to 100')

    @classmethod
    def from_json(cls, _json):
        return cls(
            assignee_id=_json.get('assigneeId', None),
            load=_json.get('load', None)
        )

    def to_json(self):
        return {
            'assigneeId': self.assignee_id,
            'load': self.load
        }


@attr.s
class Workload:
    """
    Workload object
    """

    # platform
    workload = attr.ib(type=list)

    def __iter__(self):
        for w_l_u in self.workload:
            yield w_l_u

    @workload.default
    def set_workload(self):
        workload = list()
        return workload

    @classmethod
    def from_json(cls, _json):
        return cls(
            workload=[workload_unit.from_json() for workload_unit in _json]
        )

    @staticmethod
    def _loads_are_correct(loads):
        return round(sum(loads)) == 100

    @staticmethod
    def _get_loads(num_assignees):
        loads = [0 for _ in range(num_assignees)]
        index = 0
        for i in range(10000):
            loads[index] += 1
            if index < num_assignees - 1:
                index += 1
            else:
                index = 0
        loads = [l/100 for l in loads]
        return loads

    def _redistribute(self):
        load = self._get_loads(num_assignees=len(self.workload))
        for i_w_l, w_l in self.workload:
            w_l.load = load

    @classmethod
    def generate(cls, assignee_ids, loads=None):
        if not isinstance(assignee_ids, list):
            assignee_ids = [assignee_ids]

        if loads is None:
            load = 100 / len(assignee_ids)
            loads = [load for _ in range(len(assignee_ids))]
        else:
            if not isinstance(loads, list):
                loads = [loads]
            if not Workload._loads_are_correct(loads=loads):
                raise exceptions.PlatformException('400', 'Loads must summarized to 100')

        if len(assignee_ids) != len(loads):
            raise exceptions.PlatformException('400', 'Assignee ids and loads must be of same length')

        return cls(
            workload=[WorkloadUnit(assignee_id=assignee_id, load=loads[i_assignee_id]) for i_assignee_id, assignee_id in
                      enumerate(assignee_ids)])

    def to_json(self):
        return [workload_unit.to_json() for workload_unit in self.workload]

    def add(self, assignee_id):
        self.workload.append(WorkloadUnit(assignee_id=assignee_id))
        if not self._loads_are_correct(loads=[w_l.load for w_l in self.workload]):
            self._redistribute()



import logging
import attr

from .. import entities, miscellaneous

logger = logging.getLogger("dataloop.module")


@attr.s
class PackageModule(entities.BaseEntity):
    """
    PackageModule object
    """
    # platform
    name = attr.ib(default=entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME)
    init_inputs = attr.ib()
    entry_point = attr.ib(default=entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT)
    class_name = attr.ib(default=entities.package_defaults.DEFAULT_PACKAGE_CLASS_NAME)
    functions = attr.ib()

    @functions.default
    def set_functions(self):
        functions = list()
        return functions

    @init_inputs.default
    def set_init_inputs(self):
        init_inputs = list()
        return init_inputs

    @classmethod
    def from_json(cls, _json):
        functions = [entities.PackageFunction.from_json(_function) for _function in _json.get('functions', list())]
        init_inputs = _json.get("initInputs", list())
        return cls(
            init_inputs=init_inputs,
            entry_point=_json.get("entryPoint", entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT),
            class_name=_json.get("className", entities.package_defaults.DEFAULT_PACKAGE_CLASS_NAME),
            name=_json.get("name", entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME),
            functions=functions,
        )

    def add_function(self, function):
        if not isinstance(self.functions, list):
            self.functions = [self.functions]
        if isinstance(function, entities.PackageFunction):
            self.functions.append(function)
        elif isinstance(function, dict):
            self.functions.append(entities.PackageFunction.from_json(function))
        else:
            raise ValueError('Unknown function type: {}. Expecting dl.PackageFunction or dict')

    def to_json(self):
        _json = attr.asdict(
            self,
            filter=attr.filters.exclude(attr.fields(PackageModule).functions,
                                        attr.fields(PackageModule).entry_point,
                                        attr.fields(PackageModule).class_name,
                                        attr.fields(PackageModule).init_inputs))

        # check in inputs is a list
        init_inputs = self.init_inputs
        if not isinstance(init_inputs, list):
            init_inputs = [init_inputs]
        # if is dtlpy entity convert to dict
        if init_inputs and isinstance(init_inputs[0], entities.FunctionIO):
            init_inputs = [_io.to_json() for _io in init_inputs]

        functions = self.functions
        # check in inputs is a list
        if not isinstance(functions, list):
            functions = [functions]
        # if is dtlpy entity convert to dict
        if functions and isinstance(functions[0], entities.PackageFunction):
            functions = [function.to_json() for function in functions]

        _json['entryPoint'] = self.entry_point
        _json['className'] = self.class_name
        _json['initInputs'] = init_inputs
        _json['functions'] = functions
        return _json
